#!/usr/bin/env python
# -*- coding: UTF-8 -*-
# Copyright (C) 2005-2007 Juan David Ibáñez Palomar <jdavid@itaapy.com>
# Copyright (C) 2006 Hervé Cauwelier <herve@itaapy.com>
# Copyright (C) 2007 Sylvain Taverne <sylvain@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from the Standard Library
from optparse import OptionParser
from subprocess import Popen
from sys import argv, exit
from os.path import dirname, join, realpath
from time import sleep

# Import from itools
import itools

# Import from ikaaro
from ikaaro.update import is_instance_up_to_date



def start(parser, options, target):
    # Check instance is up to date
    if not is_instance_up_to_date(target):
        print 'The instance is not up-to-date, please type:'
        print
        print '    $ icms-update.py <instance>'
        print
        return

    # Start Server
    script_path = dirname(realpath(argv[0]))
    path_icms_start_server = join(script_path, 'icms-start-server.py')
    args = [path_icms_start_server, target]
    if options.debug:
        args.append('--debug')
    if options.detach:
        args.append('--detach')
    if options.port:
        args.append('--port=%s' % options.port)
    if options.address:
        args.append('--address=%s' % options.address)
    args = ' '.join(args)
    p_server = Popen(args, shell=True)

    # Start the Mail Spool
    path_icms_start_spool = join(script_path, 'icms-start-spool.py')
    args = [path_icms_start_spool, target]
    if options.detach:
        args.append('--detach')
    args = ' '.join(args)
    p_spool = Popen(args, shell=True)

    # Detach
    if options.detach:
        sleep(2)
        exit(abs(p_server.poll() or 0)+abs(p_spool.poll() or 0))

    # Do not detach: wait for the child process
    try:
        ret = p_server.wait()
    except:
        ret = 1
    exit(abs(ret or 0)+abs(p_spool.poll() or 0))



if __name__ == '__main__':
    # The command line parser
    usage = ('%prog [OPTIONS] TARGET\n'
             '       %prog TARGET [TARGET]*')
    version = 'itools %s' % itools.__version__
    description = ('Starts a web server that publishes the TARGET ikaaro'
                   ' instance to the world. If several TARGETs are given, one'
                   ' server will be started for each one (in this mode no'
                   ' options are available).')
    parser = OptionParser(usage, version=version, description=description)
    parser.add_option(
        '-a', '--address', help='listen to IP ADDRESS')
    parser.add_option(
        '', '--debug', action="store_true", default=False,
        help="Start the server on debug mode.")
    parser.add_option(
        '-d', '--detach', action="store_true", default=False,
        help="Detach from the console.")
    parser.add_option(
        '-p', '--port', type='int', help='listen to PORT number')

    options, args = parser.parse_args()
    n_args = len(args)
    if n_args == 0:
        parser.error('The TARGET argument is missing.')
    elif n_args == 1:
        pass
    elif options.address or options.debug or options.detach or options.port:
        parser.error('Options are not available when starting several '
                     'servers at once.')

    # Action!
    for target in args:
        start(parser, options, target)



# -*- coding: UTF-8 -*-
# Copyright (C) 2007 Luis Arturo Belmar-Letelier <luis@itaapy.com>
# Copyright (C) 2007 Sylvain Taverne <sylvain@itaapy.com>
# Copyright (C) 2007-2008 Henry Obein <henry@itaapy.com>
# Copyright (C) 2007-2008 Hervé Cauwelier <herve@itaapy.com>
# Copyright (C) 2007-2008 Juan David Ibáñez Palomar <jdavid@itaapy.com>
# Copyright (C) 2007-2008 Nicolas Deram <nicolas@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from the Standard Library
from datetime import datetime, timedelta
from operator import itemgetter
from string import Template

# Import from itools
from itools.csv import Property
from itools.datatypes import Integer, String, Unicode
from itools.gettext import MSG
from itools.uri import Reference
from itools.web import ERROR
from itools.xapian import RangeQuery, AndQuery, OrQuery, PhraseQuery
from itools.xapian import StartQuery

# Import from ikaaro
from ikaaro.folder import Folder
from ikaaro.registry import register_resource_class
from resources import Resources
from stored import StoredSearch, StoredSearchFile
from tables import Tracker_TableResource, Tracker_TableHandler
from tables import ModulesResource, ModulesHandler
from tables import VersionsResource, VersionsHandler
from tracker_views import GoToIssueMenu, StoredSearchesMenu
from tracker_views import Tracker_NewInstance, Tracker_Search, Tracker_View
from tracker_views import Tracker_AddIssue, Tracker_GoToIssue
from tracker_views import Tracker_RememberSearch, Tracker_ForgetSearch
from tracker_views import Tracker_ExportToText, Tracker_ChangeSeveralBugs
from tracker_views import Tracker_ExportToCSVForm, Tracker_ExportToCSV


resolution = timedelta.resolution



default_types = [
    u'Bug', u'New Feature', u'Security Issue', u'Stability Issue',
    u'Data Corruption Issue', u'Performance Improvement',
    u'Technology Upgrade']

default_tables = [
    ('product', []),
    ('type', default_types),
    ('state', [u'Open', u'Fixed', u'Verified', u'Closed']),
    ('priority', [u'High', u'Medium', u'Low']),
    ]


class Tracker(Folder):

    class_id = 'tracker'
    class_version = '20081214'
    class_title = MSG(u'Issue Tracker')
    class_description = MSG(u'To manage bugs and tasks')
    class_icon16 = 'tracker/tracker16.png'
    class_icon48 = 'tracker/tracker48.png'
    class_views = ['search', 'add_issue', 'browse_content', 'edit']

    __fixed_handlers__ = ['product', 'module', 'version', 'type', 'priority',
        'state', 'calendar']

    @staticmethod
    def _make_resource(cls, folder, name):
        Folder._make_resource(cls, folder, name)
        # Products / Types / Priorities / States
        for table_name, values in default_tables:
            table_path = '%s/%s' % (name, table_name)
            table = Tracker_TableHandler()
            for title in values:
                title = Property(title, language='en')
                table.add_record({'title': title})
            folder.set_handler(table_path, table)
            metadata = Tracker_TableResource.build_metadata()
            folder.set_handler('%s.metadata' % table_path, metadata)
        # Modules
        table = ModulesHandler()
        folder.set_handler('%s/module' % name, table)
        metadata = ModulesResource.build_metadata()
        folder.set_handler('%s/module.metadata' % name, metadata)
        # Versions
        table = VersionsHandler()
        folder.set_handler('%s/version' % name, table)
        metadata = VersionsResource.build_metadata()
        folder.set_handler('%s/version.metadata' % name, metadata)
        # Pre-defined stored searches
        open = StoredSearchFile(state='0')
        not_assigned = StoredSearchFile(assigned_to='nobody')
        high_priority = StoredSearchFile(state='0', priority='0')
        i = 0
        for search, title in [(open, u'Open Issues'),
                              (not_assigned, u'Not Assigned'),
                              (high_priority, u'High Priority')]:
            folder.set_handler('%s/s%s' % (name, i), search)
            metadata = StoredSearch.build_metadata(title={'en': title})
            folder.set_handler('%s/s%s.metadata' % (name, i), metadata)
            i += 1
        metadata = Resources.build_metadata()
        folder.set_handler('%s/calendar.metadata' % name, metadata)


    def get_document_types(self):
        return []


    #######################################################################
    # API
    #######################################################################
    def get_new_id(self, prefix=''):
        ids = []
        for name in self.get_names():
            if prefix:
                if not name.startswith(prefix):
                    continue
                name = name[len(prefix):]
            try:
                id = int(name)
            except ValueError:
                continue
            ids.append(id)

        if ids:
            ids.sort()
            return prefix + str(ids[-1] + 1)

        return prefix + '0'


    def get_issues_query_terms(self):
        abspath = self.get_canonical_path()
        abspath = '%s/' % abspath
        return [StartQuery('abspath', abspath), PhraseQuery('format', 'issue')]


    def get_members_namespace(self, value, not_assigned=False):
        """Returns a namespace (list of dictionaries) to be used for the
        selection box of users (the 'assigned to' and 'cc' fields).
        """
        # Members
        users = self.get_resource('/users')
        members = [
            {'id': x, 'title': users.get_resource(x).get_title()}
            for x in self.get_site_root().get_members() ]
        sort_cmp = lambda x, y: cmp(x['title'].lower(), y['title'].lower())
        members.sort(cmp=sort_cmp)

        # Not assigend
        if not_assigned is True:
            members.insert(0, {'id': 'nobody', 'title': 'NOT ASSIGNED'})

        # Add 'is_selected'
        if value is None:
            condition = lambda x: False
        elif type(value) is str:
            condition = lambda x: (x == value)
        else:
            condition = lambda x: (x in value)
        for member in members:
            member['is_selected'] = condition(member['id'])

        # Ok
        return members


    def get_list_products_namespace(self):
        # Build javascript list of products/modules/versions
        products = self.get_resource('product').handler

        list_products = [{'id': '-1', 'modules': [], 'versions': []}]
        for product_record in products.get_records_in_order():
            product = {'id': product_record.id}
            for element in ['module', 'version']:
                elements = self.get_resource(element).handler

                content = []
                for record in elements.get_records_in_order():
                    product_id = elements.get_record_value(record, 'product')
                    if product_id is None:
                        continue
                    product_id = int(product_id)
                    if product_id == product_record.id:
                        content.append( {
                         'id': record.id,
                         'value': elements.get_record_value(record, 'title')})
                product[element + 's'] = content
            list_products.append(product)

        return list_products


    def get_search_results(self, context):
        """Method that return a list of issues that correspond to the search.
        """
        users = self.get_resource('/users')
        # Choose stored Search or personalized search
        search_name = context.query.get('search_name')
        if search_name:
            search = self.get_resource(search_name)
            get_value = search.handler.get_value
        else:
            get_value = context.get_query_value
        # Get search criteria
        text = get_value('text', type=Unicode)
        if text is not None:
            text = text.strip().lower()
        mtime = get_value('mtime', type=Integer)

        # Build the query
        query = self.get_issues_query_terms()
        # Text search
        if text:
            query2 = [PhraseQuery('title', text), PhraseQuery('text', text)]
            query2 = OrQuery(*query2)
            query.append(query2)
        # Metadata
        integers_type = Integer(multiple=True)
        names = 'product', 'module', 'version', 'type', 'priority', 'state'
        for name in names:
            data = get_value(name, type=integers_type)
            if len(data) > 0:
                query2 = [ PhraseQuery(name, value) for value in data ]
                query2 = OrQuery(*query2)
                query.append(query2)
        # Modification time
        if mtime:
            date = datetime.now() - timedelta(mtime)
            date = date.strftime('%Y%m%d%H%M%S')
            query2 = RangeQuery('mtime', date, None)
            query.append(query2)
        # Assign To
        assigns = get_value('assigned_to', type=String(multiple=True))
        if len(assigns) > 0:
            query2 = []
            for value in assigns:
                value = value or 'nobody'
                query2.append(PhraseQuery('assigned_to', value))
            query2 = OrQuery(*query2)
            query.append(query2)

        # Execute the search
        query = AndQuery(*query)
        return context.root.search(query)


    #######################################################################
    # User Interface
    #######################################################################
    context_menus = [GoToIssueMenu(), StoredSearchesMenu()]

    # Views
    new_instance = Tracker_NewInstance()
    search = Tracker_Search()
    view = Tracker_View()
    add_issue = Tracker_AddIssue()
    remember_search = Tracker_RememberSearch()
    forget_search = Tracker_ForgetSearch()
    go_to_issue = Tracker_GoToIssue()
    export_to_text = Tracker_ExportToText()
    export_to_csv_form = Tracker_ExportToCSVForm()
    export_to_csv = Tracker_ExportToCSV()
    change_several_bugs = Tracker_ChangeSeveralBugs()


    #######################################################################
    # Update
    #######################################################################
    def update_20080407(self):
        """Add calendar to tracker.
        """
        metadata = Resources.build_metadata()
        self.handler.set_handler('calendar.metadata', metadata)


    def update_20081015(self):
        """Add the 'products' table.
        """
        # Add the products table
        cls = Tracker_TableResource
        cls.make_resource(cls, self, 'products')
        # Change the format of the 'modules' table
        resource = self.get_resource('modules')
        metadata = resource.metadata
        metadata.set_changed()
        metadata.format = ModulesResource.class_id


    def update_20081120(self):
        """Add a default product.
        """
        from issue import Issue
        # Add a default product
        products = self.get_resource('products').get_handler()
        title = Property(u'Default', language='en')
        record = products.add_record({'title': title})
        product = record.id
        # Update Modules/Versions
        product_pro = Property(str(product))
        for name in 'modules', 'versions':
            handler = self.get_resource(name).handler
            handler.set_changed()
            handler.incremental_save = False
            for record in handler.get_records():
                for version in record:
                    version.setdefault('product', product_pro)
        # Update issues
        for issue in self.search_resources(cls=Issue):
            history = issue.get_history()
            handler.incremental_save = False
            for record in history.get_records():
                version = record[-1]
                version.setdefault('product', product)


    def update_20081214(self):
        """Rename tables.
        """
        self.move_resource('products', 'product')
        self.move_resource('modules', 'module')
        self.move_resource('versions', 'version')
        self.move_resource('types', 'type')
        self.move_resource('states', 'state')
        self.move_resource('priorities', 'priority')



###########################################################################
# Register
###########################################################################
register_resource_class(Tracker)



# -*- coding: UTF-8 -*-
# Copyright (C) 2006-2008 Hervé Cauwelier <herve@itaapy.com>
# Copyright (C) 2006-2008 Nicolas Deram <nicolas@itaapy.com>
# Copyright (C) 2007-2008 Juan David Ibáñez Palomar <jdavid@itaapy.com>
# Copyright (C) 2007-2008 Sylvain Taverne <sylvain@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from Standard Library
from calendar import monthrange, isleap
from cStringIO import StringIO
from datetime import date, datetime, time, timedelta
from operator import itemgetter

# Import from itools
from itools.csv import Property
from itools.datatypes import Date, Enumerate, Integer, String, Unicode
from itools.gettext import MSG
from itools.ical import get_grid_data
from itools.ical import Time
from itools.stl import stl
from itools.uri import encode_query, get_reference
from itools.web import BaseView, STLForm, STLView, get_context, INFO, ERROR

# Import from ikaaro
from file_views import File_Upload


resolution = timedelta.resolution

months = {
    1: MSG(u'January'),
    2: MSG(u'February'),
    3: MSG(u'March'),
    4: MSG(u'April'),
    5: MSG(u'May'),
    6: MSG(u'June'),
    7: MSG(u'July'),
    8: MSG(u'August'),
    9: MSG(u'September'),
    10: MSG(u'October'),
    11: MSG(u'November'),
    12: MSG(u'December')}

days = {
    0: MSG(u'Monday'),
    1: MSG(u'Tuesday'),
    2: MSG(u'Wednesday'),
    3: MSG(u'Thursday'),
    4: MSG(u'Friday'),
    5: MSG(u'Saturday'),
    6: MSG(u'Sunday')}



def build_timetables(start_time, end_time, interval):
    """Build a list of timetables represented as tuples(start, end).
    Interval is given by minutes.
    """
    start =  datetime(2000, 1, 1)
    if start_time:
        start = datetime.combine(start.date(), start_time)
    end =  datetime(2000, 1, 1, 23, 59)
    if end_time:
        end = datetime.combine(start.date(), end_time)

    timetables, tt_start = [], start
    while tt_start < end:
        tt_end = tt_start + timedelta(minutes=interval)
        timetables.append((tt_start.time(), tt_end.time()))
        tt_start = tt_end
    return timetables



def get_current_date(value):
    """Get date as a date object from string value.
    By default, get today's date as a date object.
    """
    if value is None:
        return date.today()
    try:
        return Date.decode(value)
    except ValueError:
        return date.today()



class Status(Enumerate):

    options = [{'name': 'TENTATIVE', 'value': MSG(u'Tentative')},
               {'name': 'CONFIRMED', 'value': MSG(u'Confirmed')},
               {'name': 'CANCELLED', 'value': MSG(u'Cancelled')}]



class TimetablesForm(STLForm):

    access = 'is_allowed_to_edit'
    title = MSG(u'Timetables')
    template = '/ui/ical/edit_timetables.xml'


    def get_namespace(self, resource, context):
        # Show current timetables only if previously set in metadata
        if resource.has_property('timetables'):
            timetables = resource.get_property('timetables')
            timetables_ns = [
                {'index': index,
                 'startname': '%s_start' % index,
                 'endname': '%s_end' % index,
                 'start': Time.encode(start),
                 'end': Time.encode(end)}
                for index, (start, end) in enumerate(timetables) ]
        else:
            timetables_ns = []

        # Ok
        return {'timetables': timetables_ns}


    action_add_schema = {'new_start': Time, 'new_end': Time}
    def action_add(self, resource, context, form):
        # Check start time is before end time
        start = form['new_start']
        end = form['new_end']
        if start >= end:
            message = ERROR(u'Start time must be earlier than end time.')
            context.message = message
            return

        # Check the given range is not defined yet
        timetables = resource.get_property('timetables')
        if (start, end) in timetables:
            context.message = ERROR(u'The given range is already defined.')
            return

        # Add new range
        timetables = list(timetables)
        timetables.append((start, end))
        timetables.sort()
        resource.set_property('timetables', tuple(timetables))
        # Ok
        context.message = INFO(u'Timetables updated successfully.')


    action_remove_schema = {'ids': Integer(multiple=True)}
    def action_remove(self, resource, context, form):
        ids = form['ids']
        if len(ids) == 0:
            context.message = ERROR(u'Nothing to remove.')
            return

        # New timetables
        timetables = resource.get_property('timetables')
        timetables = [
            timetable for index, timetable in enumerate(timetables)
            if index not in ids ]
        resource.set_property('timetables', tuple(timetables))
        # Ok
        context.message = INFO(u'Timetable(s) removed successfully.')


    def action_update(self, resource, context, form):
        timetables = resource.get_property('timetables')
        if len(timetables) == 0:
            context.message = ERROR(u'Nothing to change.')
            return

        # Update timetable or just set index to next index
        new_timetables = []
        for index in range(len(timetables)):
            try:
                start = context.get_form_value('%s_start' % index, type=Time)
                end = context.get_form_value('%s_end' % index, type=Time)
            except:
                context.message = ERROR(u'Wrong time selection (HH:MM).')
                return

            if start >= end:
                message = ERROR(u'Start time must be earlier than end time.')
                context.message = message
                return

            new_timetables.append((start, end))

        new_timetables.sort()
        resource.set_property('timetables', tuple(new_timetables))
        # Ok
        context.message = INFO(u'Timetables updated successfully.')



class CalendarView(STLView):

    # default viewed fields on monthly_view
    default_viewed_fields = ('dtstart', 'dtend', 'SUMMARY', 'STATUS')


    def get_first_day(self):
        """Returns 0 if Sunday is the first day of the week, else 1.
        For now it has to be overridden to return anything else than 1.
        """
        return 1


    def get_with_new_url(self):
        return True


    def add_selector_ns(self, c_date, method, namespace):
        """Set header used to navigate into time.

          datetime.strftime('%U') gives week number, starting week by Sunday
          datetime.strftime('%W') gives week number, starting week by Monday
          This week number is calculated as "Week O1" begins on the first
          Sunday/Monday of the year. Its range is [0,53].

        We adjust week numbers to fit rules which are used by french people.
        XXX Check for other countries
        """
        resource = get_context().resource
        if self.get_first_day() == 1:
            format = '%W'
        else:
            format = '%U'
        week_number = Unicode.encode(c_date.strftime(format))
        # Get day of 1st January, if < friday and != monday then number++
        day, kk = monthrange(c_date.year, 1)
        if day in (1, 2, 3):
            week_number = str(int(week_number) + 1)
            if len(week_number) == 1:
                week_number = '0%s' % week_number
        current_week = MSG(u'Week $n').gettext(n=week_number)
        tmp_date = c_date - timedelta(7)
        previous_week = ";%s?date=%s" % (method, Date.encode(tmp_date))
        tmp_date = c_date + timedelta(7)
        next_week = ";%s?date=%s" % (method, Date.encode(tmp_date))
        # Month
        current_month = months[c_date.month].gettext()
        delta = 31
        if c_date.month != 1:
            kk, delta = monthrange(c_date.year, c_date.month - 1)
        tmp_date = c_date - timedelta(delta)
        previous_month = ";%s?date=%s" % (method, Date.encode(tmp_date))
        kk, delta = monthrange(c_date.year, c_date.month)
        tmp_date = c_date + timedelta(delta)
        next_month = ";%s?date=%s" % (method, Date.encode(tmp_date))
        # Year
        date_before = date(c_date.year, 2, 28)
        date_after = date(c_date.year, 3, 1)
        delta = 365
        if (isleap(c_date.year - 1) and c_date <= date_before) \
          or (isleap(c_date.year) and c_date > date_before):
            delta = 366
        tmp_date = c_date - timedelta(delta)
        previous_year = ";%s?date=%s" % (method, Date.encode(tmp_date))
        delta = 365
        if (isleap(c_date.year) and c_date <= date_before) \
          or (isleap(c_date.year +1) and c_date >= date_after):
            delta = 366
        tmp_date = c_date + timedelta(delta)
        next_year = ";%s?date=%s" % (method, Date.encode(tmp_date))
        # Set value into namespace
        namespace['current_week'] = current_week
        namespace['previous_week'] = previous_week
        namespace['next_week'] = next_week
        namespace['current_month'] = current_month
        namespace['previous_month'] = previous_month
        namespace['next_month'] = next_month
        namespace['current_year'] = c_date.year
        namespace['previous_year'] = previous_year
        namespace['next_year'] = next_year
        # Add today link
        tmp_date = date.today()
        namespace['today'] = ";%s?date=%s" % (method, Date.encode(tmp_date))
        return namespace


    # Get days of week based on get_first_day's result for start
    def days_of_week_ns(self, start, num=None, ndays=7, selected=None):
        """
          start : start date of the week
          num : True if we want to get number of the day too
          ndays : number of days we want
          selected : selected date
        """
        resource = get_context().resource
        current_date = start
        ns_days = []
        for index in range(ndays):
            ns =  {}
            ns['name'] = days[current_date.weekday()].gettext()
            if num:
                ns['nday'] = current_date.day
            else:
                ns['nday'] = None
            if selected:
                ns['selected'] = (selected == current_date)
            else:
                ns['selected'] = None
            ns_days.append(ns)
            current_date = current_date + timedelta(1)
        return ns_days


    ######################################################################
    # Public API
    ######################################################################

    def get_action_url(self, **kw):
        """Action to call on form submission.
        """
        return None


    def get_events_to_display(self, start, end):
        """Get a list of events as tuples (resource_name, start, properties{})
        and a dict with all resources from whom they belong to.
        """
        resources, events = {}, []
        resource = get_context().resource
        for index, calendar in enumerate(resource.get_calendars()):
            res, evts = calendar.get_events_to_display(start, end)
            events.extend(evts)
            resources[calendar.name] = index
        events.sort(lambda x, y : cmp(x[1], y[1]))
        return resources, events


    def events_to_namespace(self, resource, events, day, cal_indexes,
                            grid=False, show_conflicts=False):
        """Build namespace for events occuring on current day.
        Update events, removing past ones.

        Events is a list of events where each one follows:
          (resource_name, dtstart, event)
          'event' object must have a methods:
              - get_end
              - get_ns_event.
        """
        ns_events = []
        index = 0
        while index < len(events):
            resource_name, dtstart, event = events[index]
            e_dtstart = dtstart.date()
            e_dtend = event.get_end().date()
            # Current event occurs on current date
            # event begins during current tt
            starts_on = e_dtstart == day
            # event ends during current tt
            ends_on = e_dtend == day
            # event begins before and ends after
            out_on = (e_dtstart < day and e_dtend > day)

            if starts_on or ends_on or out_on:
                cal_index = cal_indexes[resource_name]
                if len(cal_indexes.items()) < 2:
                    resource_name = None
                if resource_name is not None:
                    current_resource = self.get_resource(resource_name)
                else:
                    current_resource = resource
                conflicts_list = set()
                if show_conflicts:
                    handler = current_resource.handler
                    conflicts = handler.get_conflicts(e_dtstart, e_dtend)
                    if conflicts:
                        for uids in conflicts:
                            conflicts_list.update(uids)
                ns_event = event.get_ns_event(day, resource_name=resource_name,
                                              conflicts_list=conflicts_list,
                                              grid=grid, starts_on=starts_on,
                                              ends_on=ends_on, out_on=out_on)
                ns_event['url'] = current_resource.get_action_url(**ns_event)
                ns_event['cal'] = cal_index
                if 'resource' in ns_event.keys():
                    ns_event['resource']['color'] = cal_index
                else:
                    ns_event['resource'] = {'color': cal_index}
                ns_events.append(ns_event)
                # Current event end on current date
                if e_dtend == day:
                    events.remove(events[index])
                    if events == []:
                        break
                else:
                    index = index + 1
            # Current event occurs only later
            elif e_dtstart > day:
                break
            else:
                index = index + 1
        return ns_events, events



class EditEventForm(CalendarView, STLForm):

    access = 'is_allowed_to_edit'
    title = MSG(u'Edit Event')
    template = '/ui/ical/edit_event.xml'
    query_schema = {
        'resource': String,
        'id': String,
    }
    schema = {
        'SUMMARY': Unicode(mandatory=True),
        'LOCATION': Unicode,
        'dtstart': Date(mandatory=True),
        'dtstart_time': Time,
        'dtend': Date(mandatory=True),
        'dtend_time': Time,
        'DESCRIPTION': Unicode(),
        'STATUS': Status(mandatory=True),
        }


    def get_resource(self, resource, context):
        resource_id = context.query['resource']
        if resource_id is None:
            return resource

        if resource.has_resource(resource_id):
            return resource.get_resource(resource_id)
        elif resource_id == resource.name:
            return resource

        # Error
        message = MSG(u'Resource "${name}" not found.')
        context.message = message.gettext(name=resource_id)
        return None


    def get_event_id(self, resource, context):
        id = context.query['id']
        if id is not None:
            return id

        # Error
        message = ERROR(u'Expected query parameter "id" is missing.')
        context.message = message
        return None


    def get_namespace(self, resource, context):
        # Get the resource
        resource = self.get_resource(resource, context)
        if resource is None:
            return {'action': None}
        # Get the event id
        id = self.get_event_id(resource, context)
        if id is None:
            return {'action': None}
        # Get the event
        event = resource.get_record(id)
        if event is None:
            context.message = ERROR(u'Event not found')
            return {'action': None}

        # Ok
        properties = event.get_property()
        # Date start
        value = properties['DTSTART']
        value, params = value.value, value.parameters
        dtstart = Date.encode(value)
        param = params.get('VALUE', '')
        if not param or param != ['DATE']:
            dtstart_time = Time.encode(value)
        else:
            dtstart_time = None
        # Date end
        value = properties['DTEND']
        value, params = value.value, value.parameters
        dtend = Date.encode(value)
        param = params.get('VALUE', '')
        if not param or param != ['DATE']:
            dtend_time = Time.encode(value)
        else:
            dtend_time = None

        # STATUS is an enumerate
        get_value = resource.handler.get_record_value
        status = get_value(event, 'STATUS')
        status = Status().get_namespace(status)

        # Show action buttons only if current user is authorized
        if resource is None:
            allowed = True
        else:
            allowed = resource.is_organizer_or_admin(context, event)

        # The namespace
        namespace = {
            'action': ';edit_event?id=%s' % id,
            'dtstart': dtstart,
            'dtstart_time': dtstart_time,
            'dtend': dtend,
            'dtend_time': dtend_time,
            'resources': None,
            'resource': None,
            'remove': True,
            'firstday': self.get_first_day(),
            'STATUS': status,
            'allowed': allowed,
        }

        # Get values
        for key in self.schema:
            if key in namespace:
                continue
            value = properties.get(key)
            if value is None:
                namespace[key] = value
            elif isinstance(value, list):
                namespace[key] = value
            else:
                namespace[key] = value.value

        return namespace


    def get_properties(self, form):
        """Return the properties dict, ready to be used by the add or update
        actions.
        """
        # Start
        dtstart = form['dtstart']
        dtstart_time = form['dtstart_time']
        if dtstart_time is None:
            dtstart = datetime.combine(dtstart, time(0, 0))
            dtstart = Property(dtstart, VALUE=['DATE'])
        else:
            dtstart = datetime.combine(dtstart, dtstart_time)
            dtstart = Property(dtstart)
        # End
        dtend = form['dtend']
        dtend_time = form['dtend_time']
        if dtend_time is None:
            dtend = datetime.combine(dtend, time(0, 0))
            dtend = dtend + timedelta(days=1) - resolution
            dtend = Property(dtend, VALUE=['DATE'])
        else:
            dtend = datetime.combine(dtend, dtend_time)
            dtend = Property(dtend)

        # Get id and Record object
        properties = {
            'DTSTART': dtstart,
            'DTEND': dtend}

        for key in self.schema:
            if key.startswith('dtstart') or key.startswith('dtend'):
                continue
            properties[key] = Property(form[key])

        return properties


    def action_edit_event(self, resource, context, form):
        # Get the resource
        resource = self.get_resource(resource, context)
        if resource is None:
            return {'action': None}
        # Get the event id
        id = self.get_event_id(resource, context)
        if id is None:
            return {'action': None}
        # Get the event
        event = resource.get_record(id)
        if event is None:
            context.message = ERROR(u'Event not found')
            return {'action': None}

        # Test if current user is admin or organizer of this event
        if not resource.is_organizer_or_admin(context, event):
            message = ERROR(u'You are not authorized to modify this event.')
            context.message = message
            return

        # Update
        properties = self.get_properties(form)
        resource.update_record(id, properties)
        # Ok
        context.message = INFO(u'Data updated')


    def action_remove_event(self, resource, context, form):
        # Get the resource
        resource = self.get_resource(resource, context)
        if resource is None:
            return {'action': None}
        # Get the event id
        id = self.get_event_id(resource, context)
        if id is None:
            return {'action': None}
        # Get the event
        event = resource.get_record(id)
        if event is None:
            context.message = MSG(u'Event not found')
            return {'action': None}

        # Remove
        resource._remove_event(id)

        # Back
        method = context.get_cookie('method') or 'monthly_view'
        if method in dir(resource):
            goto = ';%s?%s' % (method, date.today())
        else:
            goto = '../;%s?%s' % (method, date.today())

        message = ERROR(u'Event definitely deleted.')
        return context.come_back(message, goto=goto)


    def action_cancel(self, resource, context, form):
        goto = ';%s' % context.get_cookie('method') or 'monthly_view'
        return context.come_back(None, goto)



class AddEventForm(EditEventForm):

    query_schema = {
        'resource': String,
        'date': Date,
        'start_time': Time,
        'end_time': Time,
    }


    def get_namespace(self, resource, context):
        # Get date to add event
        selected_date = context.query['date']
        if selected_date is None:
            message = u'To add an event, click on + symbol from the views.'
            context.message = ERROR(message)
            return {}

        # Timetables
        start_time = context.query['start_time']
        if start_time:
            start_time = Time.encode(start_time)
        end_time = context.query['end_time']
        if end_time:
            end_time = Time.encode(end_time)

        # The namespace
        resources = [
            {'name': x.name, 'value': x.get_title(), 'selected': False}
            for x in resource.get_calendars() ]
        namespace = {
            'action': ';add_event?date=%s' % selected_date,
            'dtstart': selected_date,
            'dtstart_time': start_time,
            'dtend': selected_date,
            'dtend_time': end_time,
            'resources': resources,
            'resource': None,
            'remove': False,
            'firstday': self.get_first_day(),
            'STATUS': Status().get_namespace(None),
            'allowed': True,
        }

        # Get values
        for key in self.schema:
            if key in namespace:
                continue
            if context.has_form_value(key):
                namespace[key] = context.get_form_value(key)
            else:
                namespace[key] = None

        return namespace


    def action_edit_event(self, resource, context, form):
        # Get the resource
        resource = self.get_resource(resource, context)
        if resource is None:
            return {'action': None}

        # Add
        properties = self.get_properties(form)
        organizer = str(context.user.get_abspath())
        properties['ORGANIZER'] = Property(organizer)
        resource.add_record('VEVENT', properties)
        # Ok
        message = INFO(u'Data updated')
        goto = ';%s' % context.get_cookie('method') or 'monthly_view'
        return context.come_back(message, goto=goto)



class MonthlyView(CalendarView):

    access = 'is_allowed_to_view'
    title = MSG(u'Monthly View')
    template = '/ui/ical/monthly_view.xml'
    monthly_template = '/ui/ical/monthly_template.xml'


    def get_namespace(self, resource, context, ndays=7):
        today_date = date.today()

        # Current date
        c_date = context.get_form_value('date')
        c_date = get_current_date(c_date)
        # Save selected date
        context.set_cookie('selected_date', c_date)

        # Method
        method = context.get_cookie('method')
        if method != 'monthly_view':
            context.set_cookie('method', 'monthly_view')

        ###################################################################
        # Calculate start of previous week
        # 0 = Monday, ..., 6 = Sunday
        weekday = c_date.weekday()
        start = c_date - timedelta(7 + weekday)
        if self.get_first_day() == 0:
            start = start - timedelta(1)
        # Calculate last date to take in account as we display  5*7 = 35 days
        end = start + timedelta(35)

        ###################################################################
        # Get a list of events to display on view
        cal_indexes, events = resource.get_events_to_display(start, end)
        if isinstance(self.monthly_template, str):
            template = resource.get_resource(self.monthly_template)
        else:
            template = self.monthly_template

        ###################################################################
        namespace = {}
        # Add header to navigate into time
        namespace = self.add_selector_ns(c_date, 'monthly_view', namespace)
        # Get header line with days of the week
        namespace['days_of_week'] = self.days_of_week_ns(start, ndays=ndays)

        namespace['weeks'] = []
        day = start
        # 5 weeks
        for w in range(5):
            ns_week = {'days': [], 'month': u''}
            # 7 days a week
            for d in range(7):
                # day in timetable
                if d < ndays:
                    ns_day = {}
                    ns_day['nday'] = day.day
                    ns_day['selected'] = (day == today_date)
                    ns_day['url'] = resource.get_action_url(day=day)
                    # Insert events
                    ns_events, events = self.events_to_namespace(resource,
                        events, day, cal_indexes)
                    ns_day['events'] = stl(template, {'events': ns_events})
                    ns_week['days'].append(ns_day)
                    if day.day == 1:
                        month = months[day.month].gettext()
                        ns_week['month'] = month
                day = day + timedelta(1)
            namespace['weeks'].append(ns_week)

        namespace['add_icon'] = '/ui/icons/16x16/add.png'
        return namespace



class WeeklyView(CalendarView):

    access = 'is_allowed_to_view'
    title = MSG(u'Weekly View')
    template = '/ui/ical/weekly_view.xml'


    def get_weekly_templates(self):
        """Get weekly templates to display events with timetables, and full
        day events.
        """
        return None, None


    # Get timetables as a list of string containing time start of each one
    def get_timetables_grid_ns(self, resource, start_date):
        """Build namespace to give as grid to gridlayout factory.
        """
        ns_timetables = []
        for calendar in resource.get_calendars():
            for start, end in calendar.get_timetables():
                for value in (start, end):
                    value = Time.encode(value)
                    if value not in ns_timetables:
                        ns_timetables.append(value)
        return ns_timetables


    def get_grid_events(self, resource, start_date, ndays=7, headers=None,
                        step=timedelta(1)):
        """Build namespace to give as data to gridlayout factory.
        """
        # Get events by day
        ns_days = []
        current_date = start_date

        if headers is None:
            headers = [None] * ndays

        # For each found calendar (or self), get events
        events = []
        # Get a list of events to display on view
        end = start_date + timedelta(days=ndays)
        cal_indexes, events = resource.get_events_to_display(start_date, end)
        for header in headers:
            ns_day = {}
            # Add header if given
            ns_day['header'] = header
            # Insert events
            ns_events, events = self.events_to_namespace(resource, events,
                                current_date, cal_indexes, grid=True)
            ns_day['events'] = ns_events
            ns_days.append(ns_day)
            current_date = current_date + step

        return ns_days


    def get_namespace(self, resource, context, ndays=7):
        # Current date
        c_date = context.get_form_value('date')
        if not c_date:
            c_date = context.get_cookie('selected_date')
        c_date = get_current_date(c_date)
        # Save selected date
        context.set_cookie('selected_date', c_date)

        # Method
        method = context.get_cookie('method')
        if method != 'weekly_view':
            context.set_cookie('method', 'weekly_view')

        # Calculate start of current week: 0 = Monday, ..., 6 = Sunday
        weekday = c_date.weekday()
        start = c_date - timedelta(weekday)
        if self.get_first_day() == 0:
            start = start - timedelta(1)

        # Add header to navigate into time
        namespace = self.add_selector_ns(c_date, 'weekly_view', {})

        # Get icon to appear to add a new event
        namespace['add_icon'] = '/ui/icons/16x16/add.png'

        # Get header line with days of the week
        days_of_week_ns = self.days_of_week_ns(start, True, ndays, c_date)
        ns_headers = []
        for day in days_of_week_ns:
            ns_header = '%s %s' % (day['name'], day['nday'])
            # Tip: Use 'selected' for css class to highlight selected date
            ns_headers.append(ns_header)
        # Calculate timetables and events occurring for current week
        timetables = self.get_timetables_grid_ns(resource, start)

        events = self.get_grid_events(resource, start, headers=ns_headers)

        # Fill data with grid (timetables) and data (events for each day)
        templates = self.get_weekly_templates()
        with_new_url = self.get_with_new_url()
        timetable = get_grid_data(events, timetables, start, templates,
                                  with_new_url)
        namespace['timetable_data'] = timetable

        return namespace



class DailyView(CalendarView):

    access = 'is_allowed_to_view'
    title = MSG(u'Daily View')
    template = '/ui/ical/daily_view.xml'
    query_schema = {
        'date': Date}


    # Start 07:00, End 21:00, Interval 30min
    class_cal_range = (time(7,0), time(21,0), 30)
    class_cal_fields = ('SUMMARY', 'DTSTART', 'DTEND')


    def get_cal_range(self):
        return self.class_cal_range


    # Get namespace for a resource's lines into daily_view
    def get_ns_calendar(self, calendar, c_date, timetables,
                        method='daily_view', show_conflicts=False):
        cal_fields = self.class_cal_fields
        calendar_name = str(calendar.name)
        args = {'date': Date.encode(c_date), 'method': method}

        # Get a dict for each event, compute colspan
        handler = calendar.handler
        events_by_index = {}
        for event in handler.search_events_in_date(c_date):
            event_start = event.get_property('DTSTART').value
            event_end = event.get_property('DTEND').value
            # Compute start and end indexes
            tt_start = 0
            tt_end = len(timetables) - 1
            for tt_index, (start, end) in enumerate(timetables):
                start = datetime.combine(c_date, start)
                end = datetime.combine(c_date, end)
                if start <= event_start:
                    tt_start = tt_index
                if end >= event_end:
                    tt_end = tt_index
                    break
            uid = getattr(event, 'id', getattr(event, 'uid', None))
            events_by_index.setdefault(tt_start, [])
            events_by_index[tt_start].append({
                'SUMMARY': event.get_property('SUMMARY').value,
                'tt_start': tt_start,
                'tt_end': tt_end,
                'resource_id': calendar_name,
                'event_id': str(uid),
                'colspan': tt_end - tt_start + 1})

        # Organize events in rows
        # If a row index is busy, start a new row
        rows = []
        for index in range(len(timetables)):
            events = events_by_index.get(index)
            if events is None:
                continue
            # Sort events by tt_end to reduce fragmentation
            # Longer events go on lines of their own
            events.sort(key=itemgetter('tt_end'))
            for row_index, event in enumerate(events):
                if not rows or len(rows) <= row_index:
                    rows.append({'events': []})
                current_events = rows[row_index]['events']
                if (current_events
                        and current_events[-1]['tt_end'] >= index):
                    # Overlapping, move on a line of its own
                    rows.append({'events': [event]})
                else:
                    # Enough free space, extend
                    current_events.append(event)

        # Get the list of conflicting events if activated
        if show_conflicts:
            conflicts_list = set()
            conflicts = handler.get_conflicts(c_date)
            if conflicts:
                for uids in conflicts:
                    uids = ['%s/%s' % (calendar_name, uid) for uid in uids]
                    conflicts_list.update(uids)

        # Organize columns
        rows_namespace = []
        for row in rows:
            row_namespace = {}
            columns_namespace = []
            events = row['events']
            event = events.pop(0)
            colspan = 0
            for tt_index, (start, end) in enumerate(timetables):
                if colspan > 0:
                    colspan = colspan - 1
                    continue
                tmp_args = args.copy()
                tmp_args['start_time'] = Time.encode(start)
                tmp_args['end_time'] = Time.encode(end)
                # Init column
                column =  {'class': None,
                           'colspan': 1,
                           'rowspan': 1,
                           'evt_url': None}
                # Add event
                if event and tt_index == event['tt_start']:
                    resource_id = event['resource_id']
                    event_id = event['event_id']
                    tmp_args = args.copy()
                    tmp_args['resource'] = resource_id
                    tmp_args['id'] = event_id
                    go_url = ';edit_event?%s' % encode_query(tmp_args)
                    if show_conflicts and uid in conflicts_list:
                        css_class = 'cal_conflict'
                    else:
                        css_class = 'cal_busy'
                    column['class'] = css_class
                    column['colspan'] = event['colspan']
                    column['evt_url'] = go_url
                    column['SUMMARY'] = event['SUMMARY']
                    # Set colspan
                    colspan = event['colspan'] - 1
                    # Delete added event
                    event = None
                    if events != []:
                        event = events.pop(0)
                # Fields in template but not shown
                for field in cal_fields:
                    if field not in column:
                        column[field] = None
                columns_namespace.append(column)
                row_namespace['columns'] = columns_namespace
            rows_namespace.append(row_namespace)

        # Header columns (one line with header and empty cases with only
        # '+' for daily_view)
        url = ';add_event?%s' % encode_query(args)
        url = get_reference(url).replace(resource=calendar_name)
        header_columns = [
            url.replace(start_time=Time.encode(x), end_time=Time.encode(y))
            for x, y in timetables ]

        # Return namespace
        return {
            'name': calendar.get_title(),
            'rows': rows_namespace,
            'header_columns': header_columns,
            'url': ';monthly_view?%s' % encode_query(args),
            'rowspan': len(rows) + 1,
        }


    def get_namespace(self, resource, context):
        method = context.get_cookie('method')
        if method != 'daily_view':
            context.set_cookie('method', 'daily_view')

        # Current date
        c_date = context.query['date']
        if c_date is None:
            c_date = date.today()

        # Add a header line with start time of each timetable
        start, end, interval = self.get_cal_range()
        timetables = build_timetables(start, end, interval)

        # Table heading and footer with the time ranges
        delta = timedelta(minutes=45)
        tt_start, tt_end = timetables[0]
        last_start = datetime.combine(c_date, tt_start)
        ns_timetables = [last_start.strftime('%H:%M')]
        # Add next ones if delta time > delta minutes
        for tt_start, tt_end in timetables[1:]:
            tt_start = datetime.combine(c_date, tt_start)
            if (tt_start - last_start) > delta:
                ns_timetables.append(tt_start.strftime('%H:%M'))
                last_start = tt_start
            else:
                ns_timetables.append(None)

        # For each found calendar
        ns_calendars = [
            self.get_ns_calendar(x, c_date, timetables)
            for x in resource.get_calendars() ]

        # Ok
        return {
            'date': Date.encode(c_date),
            'firstday': self.get_first_day(),
            'header_timetables': ns_timetables,
            'calendars': ns_calendars}



class Calendar_Upload(File_Upload):

    template = '/ui/ical/upload.xml'

    def action(self, resource, context, form):
        file = form['file']
        filename, mimetype, body = file

        # Check wether the handler is able to deal with the uploaded file
        handler = resource.handler
        if mimetype != 'text/calendar':
            message = u'Unexpected file of mimetype ${mimetype}.'
            context.message = ERROR(message, mimetype=mimetype)
            return

        # Replace
        try:
            handler.load_state_from_ical_file(StringIO(body))
        except:
            message = ERROR(u'Failed to load the file, may contain errors.')
            context.message = message
        else:
            context.server.change_resource(resource)
            context.message = INFO(u'Version uploaded')



class Calendar_Download(BaseView):

    access = 'is_allowed_to_view'

    def GET(self, resource, context):
        response = context.response
        response.set_header('Content-Type', 'text/calendar')
        # TODO Set a filename
        return resource.handler.to_ical()




# -*- coding: UTF-8 -*-
# Copyright (C) 2008 Nicolas Deram <nicolas@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from itools
from itools.datatypes import String, Tokens
from itools.gettext import MSG
from itools.handlers import merge_dicts
from itools.web import INFO
from itools.xapian import AndQuery, PhraseQuery, StartQuery
from itools.xml import XMLParser

# Import from ikaaro
from ikaaro.buttons import Button
from ikaaro.folder_views import Folder_BrowseContent
from ikaaro.forms import PathSelectorWidget
from ikaaro.registry import register_resource_class
from ikaaro.resource_views import Breadcrumb, DBResource_AddLink
from ikaaro.table import OrderedTableFile, OrderedTable
from ikaaro.table_views import OrderedTable_View
from ikaaro.views import BrowseForm, CompositeForm
from ikaaro.workflow import WorkflowAware


class AddButton(Button):

    access = 'is_allowed_to_edit'
    name = 'add'
    title = MSG(u'Add to ordered list')



###########################################################################
# ResourcesOrderedTable
###########################################################################

class ResourcesOrderedTable_Ordered(OrderedTable_View):

    title = MSG('Ordered items')

    def get_table_columns(self, resource, context):
        columns = OrderedTable_View.get_table_columns(self, resource, context)
        # Remove column with id and replace it by title,
        # and add a column for the workflow state
        columns[1] = ('title', MSG(u'Title'))
        columns.append(('workflow_state', MSG(u'State')))
        return columns


    def get_item_value(self, resource, context, item, column):
        if column == 'title':
            item = resource.get_resource_by_name(item.name, context)
            return item.get_title(), context.get_link(item)
        elif column == 'workflow_state':
            # The workflow state
            item = resource.get_resource_by_name(item.name, context)
            if not isinstance(item, WorkflowAware):
                return None
            statename = item.get_statename()
            state = item.get_state()
            msg = state['title'].gettext().encode('utf-8')
            path = context.get_link(item)
            # TODO Include the template in the base table
            state = ('<a href="%s/;edit_state" class="workflow">'
                     '<strong class="wf_%s">%s</strong>'
                     '</a>') % (path, statename, msg)
            return XMLParser(state)
        return OrderedTable_View.get_item_value(self, resource, context, item,
                                                column)


class ResourcesOrderedTable_Unordered(Folder_BrowseContent):

    access = 'is_allowed_to_edit'
    title = MSG('Unordered items')

    table_columns = [
        ('checkbox', None),
        ('title', MSG(u'Title')),
        ('path', MSG(u'Path')),
        ('workflow_state', MSG(u'State'))]

    table_actions = [AddButton]

    # Reset unrequired stuff
    context_menus = []
    search_template = None
    search_schema = {}
    def get_search_namespace(self, resource, context):
        return {}
    def get_query_schema(self):
        return {}


    def get_items(self, resource, context):
        exclude = [resource.name] + list(resource.get_ordered_names())
        orderable_classes = resource.get_orderable_classes() or ()
        path = resource.get_root_order_path(context)

        query = [StartQuery('abspath', str(path))]
        for cl in orderable_classes:
            query.append(PhraseQuery('format', cl.class_id))
        query = AndQuery(*query)
        root = context.root
        results = root.search(query).get_documents()
        items = []
        for item in results:
            if item.name not in exclude:
                item = root.get_resource(item.abspath)
                items.append(item)
        return items


    def get_item_value(self, resource, context, item, column):
        if column == 'checkbox':
            return item.name, False
        if column == 'title':
            return item.get_title(), context.get_link(item)
        if column == 'path':
            return item.name
        if column == 'workflow_state':
            # The workflow state
            if not isinstance(item, WorkflowAware):
                return None
            statename = item.get_statename()
            state = item.get_state()
            msg = state['title'].gettext().encode('utf-8')
            path = context.get_link(item)
            # TODO Include the template in the base table
            state = ('<a href="%s/;edit_state" class="workflow">'
                     '<strong class="wf_%s">%s</strong>'
                     '</a>') % (path, statename, msg)
            return XMLParser(state)


    def sort_and_batch(self, resource, context, items):
        return items


    def action_add(self, resource, context, form):
        handler = resource.handler

        order = resource.get_property('order') or []
        if not isinstance(order, list):
            order = list(order)

        index = handler.get_n_records()
        for name in form['ids']:
            handler.add_record({'name': name})
            order.append(str(index))
            index += 1

        resource.set_property('order', order)
        context.message = INFO(u'Resources added to ordered list.')



class ResourcesOrderedTable_View(CompositeForm):

    access = 'is_allowed_to_edit'
    title = MSG(u'View')
    template = '/ui/future/order_view.xml'

    subviews = [ResourcesOrderedTable_Ordered(),
                ResourcesOrderedTable_Unordered()]


    def get_namespace(self, resource, context):
        views = []
        for view in self.subviews:
            views.append({'title': view.title,
                          'view': view.GET(resource, context)})
        return {'views': views}


###########################################################################
# ChildrenOrderedTable
###########################################################################

class ChildrenOrderedTable_Ordered(ResourcesOrderedTable_Ordered):
    pass



class ChildrenOrderedTable_Unordered(ResourcesOrderedTable_Unordered):

    def get_items(self, resource, context):
        exclude = [resource.name] + list(resource.get_ordered_names())
        orderable_classes = resource.get_orderable_classes() or ()
        items = []
        for item in resource.parent.search_resources(cls=orderable_classes):
            if item.name not in exclude:
                items.append(item)
        return items



class ChildrenOrderedTable_View(ResourcesOrderedTable_View):

    subviews = [ChildrenOrderedTable_Ordered(),
                ChildrenOrderedTable_Unordered()]



class ChildrenOrderedTable_AddLink(DBResource_AddLink):

    access = 'is_allowed_to_edit'
    template = '/ui/future/order_addlink.xml'

    def get_namespace(self, resource, context):
        namespace = DBResource_AddLink.get_namespace(self, resource, context)
        exclude = [resource.name] + list(resource.get_ordered_names())
        orderable_classes = resource.get_orderable_classes() or ()

        # Construct namespace
        start = resource.parent
        bc = Breadcrumb(root=start, start=start, icon_size=48)
        items = []
        for item in bc.items:
            item_type = item['type']
            if item['name'] in exclude:
                continue
            if not issubclass(item_type, orderable_classes):
                continue
            path = str(item['path'])
            if path.startswith('../'):
                item['path'] = path[3:]
            items.append(item)
        bc.items = items
        namespace['bc'] = bc
        namespace['target_id'] = context.get_form_value('target_id')

        return namespace


###########################################################################
# Resources
###########################################################################

class ResourcesOrderedTableFile(OrderedTableFile):

    record_schema = {'name': String(mandatory=True, unique=True, index='keyword')}



class ChildrenOrderedTableFile(ResourcesOrderedTableFile):
    # FIXME 0.60 -> to remove
    record_schema = {'name': String}



class ResourcesOrderedTable(OrderedTable):

    class_id = 'resources-ordered-table'
    class_title = MSG(u'Resources Ordered Table')
    class_handler = ResourcesOrderedTableFile

    orderable_classes = None
    form = [PathSelectorWidget('name', title=MSG(u'Path'))]


    @classmethod
    def get_metadata_schema(cls):
        return merge_dicts(OrderedTable.get_metadata_schema(),
                           order=Tokens(default=()))


    def get_orderable_classes(self):
        return self.orderable_classes


    def get_root_order_path(self, context):
        """ Every item will be into this path. """
        return context.root.get_abspath()


    def get_ordered_names(self):
        for record in self.handler.get_records_in_order():
            yield record.name


    def get_resource_by_name(self, name, context):
        root_order_path = self.get_root_order_path(context)
        folder = self.get_resource(root_order_path)
        return folder.get_resource(name)


    def get_links(self):
        base = self.get_abspath()
        handler = self.handler
        links = []

        for record in handler.get_records_in_order():
            # Target resources
            path = handler.get_record_value(record, 'name')
            links.append(str(base.resolve(path)))

        return links



class ChildrenOrderedTable(ResourcesOrderedTable):

    class_id = 'children-ordered-table'
    class_title = MSG(u'Children Ordered Table')
    class_handler = ChildrenOrderedTableFile


    def get_root_order_path(self, context):
        """ Every item will be into this path. """
        return self.parent.get_abspath()


    def get_resource_by_name(self, name, context):
        return self.parent.get_resource(name)

    # Views
    add_link = ChildrenOrderedTable_AddLink()
    ordered = ChildrenOrderedTable_Ordered()
    unordered = ChildrenOrderedTable_Unordered()
    view = ChildrenOrderedTable_View()


register_resource_class(ResourcesOrderedTable)
register_resource_class(ChildrenOrderedTable)



# -*- coding: UTF-8 -*-
# Copyright (C) 2007 Henry Obein <henry@itaapy.com>
# Copyright (C) 2007 Sylvain Taverne <sylvain@itaapy.com>
# Copyright (C) 2007-2008 Hervé Cauwelier <herve@itaapy.com>
# Copyright (C) 2007-2008 Juan David Ibáñez Palomar <jdavid@itaapy.com>
# Copyright (C) 2008 Nicolas Deram <nicolas@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from itools
from itools.datatypes import String
from itools.gettext import MSG
from itools.i18n import format_datetime
from itools.vfs import FileName
from itools.web import STLForm, INFO

# Import from ikaaro
from ikaaro.folder import Folder
from ikaaro.messages import MSG_DELETE_SELECTION
from ikaaro.forms import rte_widget
from ikaaro.registry import register_resource_class
from message import Message, build_message


###########################################################################
# Views
###########################################################################
class Thread_View(STLForm):

    access = 'is_allowed_to_view'
    title = MSG(u'View')
    icon = 'view.png'
    template = '/ui/forum/Thread_view.xml'

    schema = {
        'ids': String(multiple=True, mandatory=True)
        }

    def get_namespace(self, resource, context):
        context.styles.append('/ui/forum/forum.css')
        user = context.user
        users = resource.get_resource('/users')
        ac = resource.get_access_control()
        accept = context.accept_language

        messages = []
        for message in resource.get_posts():
            author = message.get_owner()
            if author is not None:
                author = users.get_resource(author).get_title()
            messages.append({
                'name': message.name,
                'link': context.get_link(message),
                'author': author,
                'mtime': format_datetime(message.get_mtime(), accept),
                'body': message.handler.events,
            })

        # The namespace
        namespace = {
            'editable': ac.is_admin(user, resource),
            'remove_message': MSG_DELETE_SELECTION,
            'messages': messages,
            'is_allowed_to_add': ac.is_allowed_to_add(user, resource),
            }
        if namespace['is_allowed_to_add']:
            namespace['rte'] = rte_widget.to_html(String, None)
        return namespace


    action_new_reply_schema = {'data': String(mandatory=True)}
    def action_new_reply(self, resource, context, form):
        # Add
        id = resource.get_last_post_id()
        name = str(id + 1)
        data = form['data']
        language = resource.get_content_language(context)
        thread = Message.make_resource(Message, resource, name, data, language)
        # Ok
        return context.come_back(INFO(u'Reply posted'))


    def action_remove(self, resource, context, form):
        user = context.user
        for name in form['ids']:
            child = resource.get_resource(name)
            ac = child.get_access_control()
            if ac.is_allowed_to_remove(user, child):
                resource.del_resource(name)
        message = INFO(u"Message(s) deleted !")
        return context.come_back(message, goto='#new_reply')


###########################################################################
# Model
###########################################################################

class Thread(Folder):

    class_id = 'ForumThread'
    class_version = '20071215'
    class_title = MSG(u'Thread')
    class_description = u"A thread to discuss"
    class_views = ['view', 'edit']

    @staticmethod
    def _make_resource(cls, folder, name, data='', language='en'):
        Folder._make_resource(cls, folder, name)
        # First post
        folder.set_handler('%s/0.metadata' % name, Message.build_metadata())
        message = build_message(data)
        folder.set_handler('%s/0.xhtml.%s' % (name, language), message)


    def to_text(self):
        # Index the thread by the content of all its posts
        text = [ x.to_text() for x in self.search_resources(cls=Message) ]
        return u'\n'.join(text)


    def get_document_types(self):
        return [self.message_class]


    def get_posts(self):
        posts = self.search_resources(cls=Message)
        posts = [ (int(FileName.decode(x.name)[0]), x) for x in posts ]
        posts.sort()
        return [ x[1] for x in posts ]


    def get_last_post_id(self):
        ids = [ int(x.name) for x in self.search_resources(cls=Message) ]
        return max(ids)


    # Views
    view = Thread_View()


###########################################################################
# Register
###########################################################################
register_resource_class(Thread)



# -*- coding: UTF-8 -*-
# Copyright (C) 2006-2007 Hervé Cauwelier <herve@itaapy.com>
# Copyright (C) 2006-2008 Juan David Ibáñez Palomar <jdavid@itaapy.com>
# Copyright (C) 2008 Nicolas Deram <nicolas@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from itools
from itools.csv import CSVFile
from itools.gettext import POFile, MSG
from itools.handlers import TextFile, Python as PythonFile
from itools.html import HTMLFile
from itools.xml import XMLFile

# Import from ikaaro
from file import File
from file_views import File_Upload
from registry import register_resource_class
from text_views import Text_Edit, Text_View, Text_ExternalEdit, PO_Edit
from text_views import CSV_View, CSV_AddRow, CSV_EditRow



class Text(File):

    class_id = 'text'
    class_version = '20071216'
    class_title = MSG(u'Plain Text')
    class_icon16 = 'icons/16x16/text.png'
    class_icon48 = 'icons/48x48/text.png'
    class_views = ['view', 'edit', 'externaledit', 'upload',
                   'edit_state', 'history']
    class_handler = TextFile


    def get_content_type(self):
        return '%s; charset=UTF-8' % File.get_content_type(self)


    # Views
    view = Text_View()
    edit = Text_Edit()
    upload = File_Upload()
    externaledit = Text_ExternalEdit()



class PO(Text):

    class_id = 'text/x-gettext-translation'
    class_version = '20071216'
    class_title = MSG(u'Message Catalog')
    class_icon16 = 'icons/16x16/po.png'
    class_icon48 = 'icons/48x48/po.png'
    class_handler = POFile

    # Views
    edit = PO_Edit()



class CSS(Text):

    class_id = 'text/css'
    class_version = '20071216'
    class_title = MSG(u'CSS')
    class_icon16 = 'icons/16x16/css.png'
    class_icon48 = 'icons/48x48/css.png'



class Python(Text):

    class_id = 'text/x-python'
    class_version = '20071216'
    class_title = MSG(u'Python')
    class_icon16 = 'icons/16x16/python.png'
    class_icon48 = 'icons/48x48/python.png'
    class_handler = PythonFile



class XML(Text):

    class_id = 'text/xml'
    class_version = '20071216'
    class_title = MSG(u'XML File')
    class_handler = XMLFile



class HTML(Text):

    class_id = 'text/html'
    class_version = '20071216'
    class_title = MSG(u'HTML File')
    class_handler = HTMLFile



class CSV(Text):

    class_id = 'text/comma-separated-values'
    class_version = '20071216'
    class_title = MSG(u'Comma Separated Values')
    class_views = ['view', 'add_row', 'edit', 'externaledit', 'upload',
                   'history']
    class_handler = CSVFile


    def get_columns(self):
        """Returns a list of tuples with the name and title of every column.
        """
        handler = self.handler

        if handler.columns is None:
            row = None
            for row in handler.lines:
                if row is not None:
                    break
            if row is None:
                return []
            return [ (str(x), str(x)) for x in range(len(row)) ]

        columns = []
        for name in handler.columns:
            datatype = handler.schema[name]
            title = getattr(datatype, 'title', None)
            if title is None:
                title = name
            else:
                title = title.gettext()
            columns.append((name, title))

        return columns


    # Views
    edit = None
    view = CSV_View()
    add_row = CSV_AddRow()
    edit_row = CSV_EditRow()



###########################################################################
# Register
###########################################################################
register_resource_class(Text)
register_resource_class(Python)
register_resource_class(PO)
register_resource_class(CSS)
register_resource_class(XML)
register_resource_class(XML, format='application/xml')
register_resource_class(HTML)
register_resource_class(CSV)
register_resource_class(CSV, 'text/x-comma-separated-values')
register_resource_class(CSV, 'text/csv')




# -*- coding: UTF-8 -*-
# Copyright (C) 2005-2008 Juan David Ibáñez Palomar <jdavid@itaapy.com>
# Copyright (C) 2006-2007 Hervé Cauwelier <herve@itaapy.com>
# Copyright (C) 2006-2007 Nicolas Deram <nicolas@itaapy.com>
# Copyright (C) 2007 Henry Obein <henry@itaapy.com>
# Copyright (C) 2007 Sylvain Taverne <sylvain@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from the Standard Library
from copy import deepcopy

# Import from itools
from itools import get_abspath
from itools.gettext import MSG
from itools.handlers import File, Folder, Image
from itools.http import NotFound
from itools.i18n import has_language
from itools.stl import stl
from itools.web import get_context, BaseView, ERROR
from itools.xml import XMLFile

# Import from ikaaro
from resource_ import IResource
from skins_views import LanguagesTemplate, LocationTemplate



class FileGET(BaseView):

    access = True


    def get_mtime(self, resource):
        return resource.get_mtime()


    def GET(self, resource, context):
        response = context.response
        response.set_header('Content-Type', resource.get_mimetype())
        return resource.to_str()



class UIFile(IResource, File):

    def clone(self, cls=None, exclude=('database', 'uri', 'timestamp',
                                       'dirty', 'parent', 'name')):
        return File.clone(self, cls=cls, exclude=exclude)


    download = FileGET()

    def get_view(self, name, query=None):
        if name is None:
            return self.download
        raise NotFound



class UIImage(Image, UIFile):
    pass



class UITemplate(UIFile, XMLFile):
    pass



map = {
    'application/xml': UITemplate,
    'application/xhtml+xml': UITemplate,
    'text/xml': UITemplate,
    'image/png': UIImage,
    'image/jpeg': UIImage,
}



class UIFolder(IResource, Folder):

    class_title = MSG(u'UI')
    class_icon48 = 'icons/48x48/folder.png'


    def _get_names(self):
        # FIXME May not be the right implementation
        return self.get_handler_names()


    def _get_resource(self, name):
        if self.has_handler(name):
            handler = self.get_handler(name)
        else:
            n = len(name)
            names = [ x for x in self.get_handler_names() if x[:n] == name ]
            languages = [ x.split('.')[-1] for x in names ]
            languages = [ x for x in languages if has_language(x) ]

            if not languages:
                raise LookupError, 'resource "%s" not found' % name

            # Get the best variant
            context = get_context()
            if context is None:
                language = None
            else:
                accept = context.accept_language
                language = accept.select_language(languages)

            # By default use whatever variant
            # (XXX we need a way to define the default)
            if language is None:
                language = languages[0]
            handler = self.get_handler('%s.%s' % (name, language))

        if isinstance(handler, Folder):
            handler = UIFolder(handler.uri)
        else:
            format = handler.get_mimetype()
            cls = map.get(format, UIFile)
            handler = cls(handler.uri)
        handler.database = self.database
        return handler



class Skin(UIFolder):

    class_title = MSG(u'Skin')
    class_icon16 = 'icons/16x16/skin.png'
    class_icon48 = 'icons/48x48/skin.png'

    # User Interface widgets
    languages_template = LanguagesTemplate
    location_template = LocationTemplate


    #######################################################################
    # HTML head
    #######################################################################
    def get_template_title(self, context):
        """Return the title to give to the template document.
        """
        here = context.resource
        # In the Root
        root = here.get_site_root()
        if root is here:
            return root.get_title()
        # Somewhere else
        message = MSG(u"${root_title}: ${here_title}")
        return message.gettext(root_title=root.get_title(),
                               here_title=here.get_title())


    def get_styles(self, context):
        styles = [
            # BackOffice style
            '/ui/bo.css',
            # Calendar JS Widget (http://dynarch.com/mishoo/calendar.epl)
            '/ui/calendar/calendar-aruni.css',
            # Calendar
            '/ui/ical/calendar.css',
            # Table
            '/ui/table/style.css']

        # This skin's style
        if self.has_handler('style.css'):
            styles.append('%s/style.css' % self.get_abspath())
        # Dynamic styles
        for style in context.styles:
            styles.append(style)

        return styles


    def get_scripts(self, context):
        scripts = []
        # Aruni (default skin)
        scripts.append('/ui/jquery.js')
        scripts.append('/ui/javascript.js')
        # Calendar (http://dynarch.com/mishoo/calendar.epl)
        scripts.append('/ui/calendar/calendar.js')
        languages = [
            'af', 'al', 'bg', 'br', 'ca', 'da', 'de', 'du', 'el', 'en', 'es',
            'fi', 'fr', 'hr', 'hu', 'it', 'jp', 'ko', 'lt', 'lv', 'nl', 'no',
            'pl', 'pt', 'ro', 'ru', 'si', 'sk', 'sp', 'sv', 'tr', 'zh']
        accept = context.accept_language
        language = accept.select_language(languages)
        scripts.append('/ui/calendar/lang/calendar-%s.js' % language)
        scripts.append('/ui/calendar/calendar-setup.js')
        # Table
        scripts.append('/ui/table/javascript.js')
        # This skin's JavaScript
        if self.has_handler('javascript.js'):
            scripts.append('%s/javascript.js' % self.get_abspath())
        # Dynamic scripts
        for script in context.scripts:
            scripts.append(script)

        return scripts


    def get_meta_tags(self, context):
        """Return a list of dict with meta tags to give to the template
        document.
        """
        here = context.resource
        root = here.get_site_root()

        meta = []
        # Set description
        value, language = here.get_property_and_language('description')
        if value:
            meta.append({'name': 'description', 'lang': language,
                         'content': value})
        # Set keywords for all languages
        for language in root.get_property('website_languages'):
            value = here.get_property('subject', language)
            if value is None:
                continue
            value = value.strip()
            if value:
                meta.append({'name': 'keywords', 'lang': language,
                             'content': value})
        return meta


    #######################################################################
    # Authenticated user
    #######################################################################
    def get_user_menu(self, context):
        """Return a dict {user_icon, user, joinisopen}.
        """
        user = context.user

        if user is None:
            site_root = context.site_root
            joinisopen = site_root.is_allowed_to_register(user, site_root)
            return {'info': None, 'joinisopen': joinisopen}

        home = '/users/%s' % user.name
        info = {'name': user.name, 'title': user.get_title(),
                'home': home}
        return {'info': info, 'joinisopen': False}


    #######################################################################
    # Body
    #######################################################################
    def _get_page_title(self, context):
        resource = context.resource
        view = context.view

        # Page title
        try:
            get_page_title = view.get_page_title
        except AttributeError:
            return resource.get_page_title()

        return get_page_title(resource, context)


    def get_messages(self, context):
        """Return the message string of the last action.
        A list of messages is supported.
        """
        # Text
        if context.message is not None:
            messages = context.message
        elif 'message' in context.uri.query:
            # TODO Do not use this anymore
            messages = context.get_query_value('message')
        else:
            return None

        # Multiple messages:
        if not isinstance(messages, list):
            messages = [messages]

        messages_ns = []
        for message in messages:
            css_class = 'error' if isinstance(message, ERROR) else 'info'
            messages_ns.append({'message': message, 'class': css_class})

        namespace = {'messages': messages_ns}

        template = context.root.get_resource('/ui/aruni/message.xml')
        return stl(template, namespace)


    def _get_context_menus(self, context):
        resource = context.resource
        # Resource
        for menu in resource.get_context_menus():
            yield menu.render(resource, context)
        # View
        menus = getattr(context.view, 'context_menus', [])
        for menu in menus:
            yield menu.render(resource, context)


    #######################################################################
    # Main
    #######################################################################
    def build_namespace(self, context):
        context_menus = self._get_context_menus(context)
        context_menus = list(context_menus)

        # The document language
        here = context.resource
        languages = here.get_site_root().get_property('website_languages')
        language = context.accept_language.select_language(languages)

        # The base URI
        uri = context.uri
        if uri.path and not context.view_name and not uri.path.endswith_slash:
            uri = deepcopy(uri)
            uri.path.endswith_slash = True

        # In case of UI objects, fallback to site root
        if isinstance(here, (UIFile, UIFolder)):
            base_path = ''
        else:
            base_path = context.get_link(here)

        return {
            # HTML head
            'language': language,
            'title': self.get_template_title(context),
            'base_uri': str(uri),
            'styles': self.get_styles(context),
            'scripts': self.get_scripts(context),
            'meta_tags': self.get_meta_tags(context),
            # Log in/out
            'login': '%s/;login' % base_path,
            'logout': '%s/;logout' % base_path,
            # User
            'user': self.get_user_menu(context),
            # Location & Views
            'location': self.location_template(context),
            'languages': self.languages_template(context),
            # Body
            'page_title': self._get_page_title(context),
            'message': self.get_messages(context),
            'context_menus': context_menus,
        }


    def get_template(self):
        try:
            return self.get_resource('template.xhtml')
        except LookupError:
            # Default, aruni
            return self.get_resource('/ui/aruni/template.xhtml')


    def template(self, content):
        context = get_context()
        # Build the namespace
        namespace = self.build_namespace(context)
        namespace['body'] = content

        # Set the encoding to UTF-8
        context.response.set_header('Content-Type', 'text/html; charset=UTF-8')

        # Load the template
        handler = self.get_template()

        # Build the output
        s = ['<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN"\n'
             '  "http://www.w3.org/TR/html4/strict.dtd">']
        # STL
        prefix = handler.get_abspath()
        data = stl(handler, namespace, prefix=prefix, mode='html')
        s.append(data)

        return ''.join(s)



#############################################################################
# The folder "/ui"
#############################################################################

skin_registry = {}
def register_skin(name, skin):
    if isinstance(skin, str):
        skin = Skin(skin)
    skin_registry[name] = skin


# Register the built-in skins
ui_path = get_abspath('ui')
register_skin('aruni', '%s/aruni' % ui_path)


class UI(UIFolder):

    def _get_resource(self, name):
        if name in skin_registry:
            skin = skin_registry[name]
            skin.database = self.database
            return skin
        return UIFolder._get_resource(self, name)




#!/usr/bin/env python
# -*- coding: UTF-8 -*-
# Copyright (C) 2005-2008 Juan David Ibáñez Palomar <jdavid@itaapy.com>
# Copyright (C) 2006-2007 Hervé Cauwelier <herve@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from the Standard Library
from optparse import OptionParser
import sys
from traceback import print_exc

# Import from itools
import itools
from itools.http import Request
from itools import vfs
from itools.web import set_context, Context

# Import from ikaaro
from ikaaro.resource_ import DBResource
from ikaaro.server import Server, ask_confirmation


def update(parser, options, target):
    folder = vfs.open(target)
    confirm = options.confirm

    #######################################################################
    # STAGE 0: Specific upgrades
    #######################################################################

    #######################################################################
    # STAGE 1: Find out the versions to upgrade
    #######################################################################
    print 'Please wait while we find out the versions to upgrade.'

    # Build the server object
    server = Server(target)
    database = server.database
    root = server.root

    # Find out the versions to upgrade
    versions = set()
    for resource in root.traverse_resources():
        # Skip non-database resources
        if not isinstance(resource, DBResource):
            continue

        # Skip up-to-date resources
        obj_version = resource.metadata.version
        cls_version = resource.class_version
        if obj_version == cls_version:
            continue

        # Check for code that is older than the instance
        if obj_version > cls_version:
            print
            print '*'
            print '* ERROR: resource is newer than its class'
            print '* %s <%s>' % (resource.get_abspath(), resource.__class__.__name__)
            print '* %s > %s' % (obj_version, cls_version)
            print '*'
            return

        next_versions = resource.get_next_versions()
        if not next_versions:
            continue

        sys.stdout.write('.')
        sys.stdout.flush()
        for version in next_versions:
            versions.add(version)

    versions = list(versions)
    versions.sort()

    if not versions:
        print '*'
        print '* The instance is up-to-date: nothing to do.'
        print '*'
        return

    print
    print '*'
    print '* Versions to upgrade: %s' % ', '.join(versions)
    print '*'

    #######################################################################
    # STAGE 2: General Upgrade code
    #######################################################################
    # Build a fake context
    context = Context(Request())
    context.server = server
    set_context(context)

    # Open the update log
    log = open('%s/log/update' % target, 'w')

    # Update
    for version in versions:
        message = 'Upgrade to version %s (y/N)? ' % version
        if ask_confirmation(message, confirm) is False:
            print '*'
            print '* WARNING: Upgrade process not finished.'
            print '*'
            return
        # Go ahead
        bad = 0
        for resource in root.traverse_resources():
            # Skip non-database resources
            if not isinstance(resource, DBResource):
                continue

            # Skip up-to-date resources
            obj_version = resource.metadata.version
            cls_version = resource.class_version
            if obj_version == cls_version:
                continue

            next_versions = resource.get_next_versions()
            if not next_versions:
                continue

            if next_versions[0] != version:
                continue

            # Update
            sys.stdout.write('.')
            sys.stdout.flush()
            try:
                resource.update(version)
                database.save_changes()
            except:
                path = resource.get_abspath()
                log.write('%s %s\n' % (path, resource.__class__))
                print_exc(file=log)
                log.write('\n')
                bad += 1
        # Reset the state
        database.cache = {}
        # Stop if there were errors
        print
        if bad > 0:
            print '*'
            print '* ERROR: %s resources failed to upgrade to version %s' \
                  % (bad, version)
            print '* Check the "%s/log/update" file.' % target
            print '*'
            return

    # It is Done
    print '*'
    print '* To finish the upgrade process update the catalog:'
    print '*'
    print '*   $ icms-update-catalog.py %s' % target
    print '*'



if __name__ == '__main__':
    # The command line parser
    usage = '%prog [OPTIONS] TARGET'
    version = 'itools %s' % itools.__version__
    description = ('Updates the TARGET ikaaro instance (if needed). Use'
                   ' this command when upgrading to a new version of itools.')
    parser = OptionParser(usage, version=version, description=description)
    parser.add_option(
        '-y', '--yes', action='store_true', dest='confirm',
        help="start the update without asking confirmation")

    # TODO Add option --pretend (to know whether the database needs to be
    # updated)

    options, args = parser.parse_args()
    if len(args) != 1:
        parser.error('incorrect number of arguments')

    target = args[0]

    # Action!
    update(parser, options, target)



# -*- coding: UTF-8 -*-
# Copyright (C) 2007-2008 Nicolas Deram <nicolas@itaapy.com>
# Copyright (C) 2008 Gautier Hayoun <gautier.hayoun@itaapy.com>
# Copyright (C) 2008 Juan David Ibáñez Palomar <jdavid@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from the Standard Library
from datetime import time, timedelta

# Import from itools
from itools.csv import Table as BaseTable, Record
from itools.datatypes import DateTime, String, Unicode, Enumerate
from itools.gettext import MSG
from itools.ical import Time
from itools.utils import freeze
from itools.web import get_context
from itools.xapian import OrQuery, AndQuery, RangeQuery

# Import from ikaaro
from ikaaro.calendar_ import CalendarBase
from ikaaro.calendar_views import MonthlyView
from ikaaro.calendar_views import WeeklyView
from ikaaro.forms import DateWidget, MultilineWidget, SelectWidget, TextWidget
from ikaaro.table import Table
from ikaaro.registry import register_resource_class

# Import from ikaaro.tracker
from tracker_views import Tracker_View


resolution = timedelta.resolution


class TrackerMonthlyView(MonthlyView):

    monthly_template = '/ui/tracker/monthly_template.xml'
    query_schema = Tracker_View.tracker_schema

    def get_with_new_url(self):
        return False


    def get_namespace(self, resource, context):
        context.styles.append('/ui/tracker/tracker.css')
        return MonthlyView.get_namespace(self, resource, context)



class TrackerWeeklyView(WeeklyView):

    weekly_template_fd = '/ui/tracker/weekly_template_fd.xml'
    weekly_template = '/ui/tracker/weekly_template.xml'
    query_schema = Tracker_View.tracker_schema


    def get_weekly_templates(self):
        root = get_context().root
        template = root.get_resource(self.weekly_template)
        template_fd = root.get_resource(self.weekly_template_fd)
        return template, template_fd


    def get_with_new_url(self):
        return False


    # Get timetables as a list of string containing time start of each one
    def get_timetables_grid_ns(self, resource, start_date):
        """Build namespace to give as grid to gridlayout factory.
        """
        ns_timetables = []
        for start, end in resource.get_timetables():
            for value in (start, end):
                value = Time.encode(value)
                if value not in ns_timetables:
                    ns_timetables.append(value)
        return ns_timetables


    def get_namespace(self, resource, context):
        context.styles.append('/ui/tracker/tracker.css')
        return WeeklyView.get_namespace(self, resource, context)



class Resource(Record):

    def get_end(self):
        return self.get_value('dtend')


    def get_ns_event(self, day, resource_name=None, conflicts_list=freeze([]),
                     timetable=None, grid=False, starts_on=True, ends_on=True,
                     out_on=True):
        context = get_context()
        tracker = context.resource.parent
        issue = tracker.get_resource(self.get_value('issue'))
        comment = self.get_value('comment')

        users = context.root.get_resource('/users')
        user = self.get_value('resource')
        user_title = users.get_resource(user).get_title()

        products = tracker.get_resource('product').handler
        product = issue.get_value('product')
        product = products.get_record(product)
        product = products.get_record_value(product, 'title')

        ns = {}
        ns['resource'] = {'name': user, 'title': user_title}

        ns['issue'] = {'number': issue.name, 'title': issue.get_value('title'),
                       'url': '../%s/;edit' % issue.name, 'product': product,
                       'comment': comment}

        ###############################################################
        # Set dtstart and dtend values using '...' for events which
        # appear into more than one cell
        start = self.get_value('dtstart')
        end = self.get_value('dtend')
        ns['start'] = Time.encode(start.time())
        ns['end'] = Time.encode(end.time())
        ns['TIME'] = None
        if grid:
            # Neither a full day event nor a multiple days event
            if start.time() != time(0,0) and end.time() != time(0,0)\
              and start.date() == end.date():
                ns['TIME'] = '%s - %s' % (ns['start'], ns['end'])
            else:
                ns['start'] = ns['end'] = None
        elif not out_on:
            if start.date() != end.date():
                value = ''
                if starts_on:
                    value = ns['start']
                    if ends_on:
                        value = value + '-'
                    else:
                        value = value + '...'
                if ends_on:
                    value = value + ns['end']
                    if not starts_on:
                        value = '...' + value
                ns['TIME'] = '(' + value + ')'

        return ns



class ListOfUsers(Enumerate):

    @classmethod
    def get_options(cls):
        tracker = get_context().resource.parent
        return [{'name': x['id'], 'value': x['title']}
                for x in tracker.get_members_namespace('')]



class BaseResources(BaseTable):

    record_schema = {
        'dtstart': DateTime(mandatory=True, index='keyword'),
        'dtend': DateTime(mandatory=True, index='keyword'),
        'issue': String(mandatory=True, index='keyword'),
        'resource': ListOfUsers(mandatory=True, index='keyword'),
        'comment': Unicode}

    form = [
        SelectWidget('resource', title=MSG(u'Resource')),
        DateWidget('dtstart', title=MSG(u'Start YYYY-MM-DD HH:MM')),
        DateWidget('dtend', title=MSG(u'End')),
        MultilineWidget('comment', title=MSG(u'Comment')),
        TextWidget('issue', title=MSG(u'Issue'))]

    record_class = Resource



class Resources(Table, CalendarBase):

    class_id = 'tracker_calendar'
    class_version = '20071216'
    class_title = MSG(u'Resources')
    class_description = MSG(u'Resources assigned to issues')
    class_handler = BaseResources

    class_views = ['weekly_view', 'monthly_view', 'edit_timetables']


    def get_action_url(self, **kw):
        issue = kw.get('issue', None)
        if issue:
            return get_context().uri.resolve('../%s/;edit' % issue)
        return None


    def get_events_to_display(self, start, end):
        # Build the query
        dtstart = str(start)
        dtend = str(end)
        dtstart_limit = str(start + resolution)
        dtend_limit = str(end + resolution)
        query = OrQuery(RangeQuery('dtstart', dtstart, dtend),
                        RangeQuery('dtend', dtstart_limit, dtend_limit),
                        AndQuery(RangeQuery('dtstart', None, dtstart),
                                 RangeQuery('dtend', dtend, None)))

        # Define the set of concerned issues
        tracker = self.parent
        issues = tracker.get_search_results(get_context())
        issues = [ x.name for x in issues.get_documents() ]

        # Find out the events
        handler = self.handler
        events = [
            (0, handler.get_record_value(x, 'dtstart'), x)
            for x in handler.search(query)
            if handler.get_record_value(x, 'issue') in issues ]
        events.sort(lambda x, y: cmp(x[1], y[1]))

        # Ok
        return {0: self.name}, events


    monthly_view = TrackerMonthlyView()
    weekly_view = TrackerWeeklyView()




###########################################################################
# Register
###########################################################################
register_resource_class(Resources)



# -*- coding: UTF-8 -*-
# Copyright (C) 2008 Juan David Ibáñez Palomar <jdavid@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from the Standard Library
from urllib import quote
from datetime import datetime

# Import from the Python Image Library
try:
    from PIL import Image as PILImage
except ImportError:
    PILImage = None

# Import from itools
from itools.datatypes import Boolean, Integer, String, Unicode
from itools.gettext import MSG
from itools.handlers import checkid, merge_dicts
from itools.i18n import format_datetime
from itools.stl import set_prefix
from itools.uri import get_reference, Path
from itools.web import BaseView, STLForm, ERROR
from itools.xapian import AndQuery, OrQuery, PhraseQuery
from itools.xml import XMLParser

# Import from ikaaro
from buttons import RemoveButton, RenameButton, CopyButton, CutButton
from buttons import PasteButton, PublishButton
from datatypes import CopyCookie, ImageWidth
from exceptions import ConsistencyError
import messages
from resource_views import AddResourceMenu
from utils import generate_name, get_base_path_query
from versioning import VersioningAware
from views import IconsView, SearchForm, ContextMenu
from workflow import WorkflowAware



class ZoomMenu(ContextMenu):

    title = MSG(u'Zoom')

    def get_items(self, resource, context):
        uri = context.uri

        # Compute previous and next sizes
        current_size = context.query['size']
        min_size = resource.MIN_SIZE
        max_size = resource.MAX_SIZE
        current_size = max(min_size, min(current_size, max_size))
        previous_size = min_size
        next_size = max_size
        for step in resource.SIZE_STEPS:
            if step < current_size:
                previous_size = step
            if next_size is max_size and step > current_size:
                next_size = step

        next_size = str(next_size)
        previous_size = str(previous_size)
        return [
            {'title': MSG(u'Zoom In'),
             'src': '/ui/icons/16x16/zoom_in.png',
             'href': uri.replace(size=next_size)},
            {'title': MSG(u'Zoom Out'),
             'src': '/ui/icons/16x16/zoom_out.png',
             'href': uri.replace(size=previous_size)}
        ]



class Folder_View(BaseView):

    access = 'is_allowed_to_view_folder'
    title = MSG(u'View')


    def GET(self, resource, context):
        try:
            index = resource.get_resource('index')
        except LookupError:
            context.message = ERROR(
                u'There is not an "index" web page. Could not render this '
                u'view.')
            return ''

        # Rewrite the URLs
        stream = index.get_html_data()
        return set_prefix(stream, 'index/')



class Folder_NewResource(IconsView):

    access = 'is_allowed_to_add'
    title = MSG(u'Add resource')
    icon = 'new.png'


    def get_namespace(self, resource, context):
        items = [
            {'icon': '/ui/' + cls.class_icon48,
             'title': cls.class_title.gettext(),
             'description': cls.class_description.gettext(),
             'url': ';new_resource?type=%s' % quote(cls.class_id)
            }
            for cls in resource.get_document_types() ]

        return {
            'batch': None,
            'items': items}



class Folder_Rename(STLForm):

    access = 'is_allowed_to_edit'
    title = MSG(u'Rename resources')
    template = '/ui/folder/rename.xml'
    query_schema = {
        'ids': String(multiple=True)}
    schema = {
        'paths': String(multiple=True, mandatory=True),
        'new_names': String(multiple=True, mandatory=True)}


    def get_namespace(self, resource, context):
        ids = context.query['ids']
        # Filter names which the authenticated user is not allowed to move
        ac = resource.get_access_control()
        user = context.user
        paths = [ x for x in ids
                  if ac.is_allowed_to_move(user, resource.get_resource(x)) ]

        # Build the namespace
        paths.sort()
        paths.reverse()
        items = []
        for path in paths:
            if '/' in path:
                parent_path, name = path.rsplit('/', 1)
                parent_path += '/'
            else:
                parent_path = ''
                name = path
            items.append({
                'path': path,
                'parent_path': parent_path,
                'name': name})

        return {'items': items}


    def action(self, resource, context, form):
        paths = form['paths']
        new_names = form['new_names']

        paths.sort()
        paths.reverse()
        # Clean the copy cookie if needed
        cut, cp_paths = context.get_cookie('ikaaro_cp', type=CopyCookie)

        renamed = []
        referenced = []
        # Process input data
        abspath = resource.get_abspath()
        for i, path in enumerate(paths):
            new_name = new_names[i]
            new_name = checkid(new_name)
            if new_name is None:
                context.message = messages.MSG_BAD_NAME
                return
            # Split the path
            if '/' in path:
                parent_path, old_name = path.rsplit('/', 1)
                container = resource.get_resource(parent_path)
            else:
                old_name = path
                container = resource
            # Check the name really changed
            if new_name == old_name:
                continue
            # Check there is not another resource with the same name
            if container.has_resource(new_name):
                context.message = messages.MSG_EXISTANT_FILENAME
                return
            # Clean cookie (FIXME Do not clean the cookie, update it)
            if cp_paths and str(abspath.resolve2(path)) in cp_paths:
                context.del_cookie('ikaaro_cp')
                cp_paths = []
            # Rename
            try:
                container.move_resource(old_name, new_name)
            except ConsistencyError:
                referenced.append(old_name)
                continue
            else:
                renamed.append(old_name)

        if referenced and not renamed:
            resources = ', '.join(referenced)
            message = messages.MSG_RESOURCES_REFERENCED(resources=resources)
        else:
            message = messages.MSG_RENAMED
        return context.come_back(message, goto=';browse_content')



class Folder_BrowseContent(SearchForm):

    access = 'is_allowed_to_view'
    title = MSG(u'Browse Content')
    context_menus = [AddResourceMenu()]
    schema = {
        'ids': String(multiple=True, mandatory=True),
    }

    # Search Form
    search_template = '/ui/folder/browse_search.xml'
    search_schema = {
        'search_field': String,
        'search_term': Unicode,
        'search_subfolders': Boolean(default=False),
    }

    # Table
    table_columns = [
        ('checkbox', None),
        ('icon', None),
        ('name', MSG(u'Name')),
        ('title', MSG(u'Title')),
        ('format', MSG(u'Type')),
        ('mtime', MSG(u'Last Modified')),
        ('last_author', MSG(u'Last Author')),
        ('size', MSG(u'Size')),
        ('workflow_state', MSG(u'State'))]


    def get_search_namespace(self, resource, context):
        namespace = SearchForm.get_search_namespace(self, resource, context)
        namespace['search_subfolders'] = context.query['search_subfolders']
        return namespace


    def get_items(self, resource, context, *args):
        # Get the parameters from the query
        query = context.query
        search_term = query['search_term'].strip()
        field = query['search_field']
        search_subfolders = query['search_subfolders']

        # Build the query
        args = list(args)
        abspath = str(resource.get_canonical_path())
        if search_subfolders is True:
            args.append(get_base_path_query(abspath))
        else:
            args.append(PhraseQuery('parent_path', abspath))
        if search_term:
            args.append(PhraseQuery(field, search_term))
        if len(args) == 1:
            query = args[0]
        else:
            query = AndQuery(*args)

        # Ok
        return context.root.search(query)


    def sort_and_batch(self, resource, context, results):
        start = context.query['batch_start']
        size = context.query['batch_size']
        sort_by = context.query['sort_by']
        reverse = context.query['reverse']
        items = results.get_documents(sort_by=sort_by, reverse=reverse,
                                      start=start, size=size)

        # Access Control (FIXME this should be done before batch)
        user = context.user
        root = context.root
        allowed_items = []
        for item in items:
            item = root.get_resource(item.abspath)
            ac = item.get_access_control()
            if ac.is_allowed_to_view(user, item):
                allowed_items.append(item)

        return allowed_items


    def get_item_value(self, resource, context, item, column):
        if column == 'checkbox':
            # checkbox
            parent = item.parent
            if parent is None:
                return None
            if item.name in parent.__fixed_handlers__:
                return None
            id = resource.get_canonical_path().get_pathto(item.get_abspath())
            id = str(id)
            return id, False
        elif column == 'icon':
            # icon
            path_to_icon = item.get_resource_icon(16)
            if path_to_icon.startswith(';'):
                path_to_icon = Path('%s/' % item.name).resolve(path_to_icon)
            return path_to_icon
        elif column == 'name':
            # Name
            id = resource.get_canonical_path().get_pathto(item.get_abspath())
            id = str(id)
            view = item.get_view(None)
            if view is None:
                return id
            href = '%s/' % context.get_link(item)
            return id, href
        elif column == 'title':
            # Title
            return item.get_property('title')
        elif column == 'format':
            # Type
            return item.class_title.gettext()
        elif column == 'mtime':
            # Last Modified
            accept = context.accept_language
            return format_datetime(item.get_mtime(), accept=accept)
        elif column == 'last_author':
            # Last author
            if not isinstance(item, VersioningAware):
                return None
            revisions = item.get_revisions(context)
            if not revisions:
                return None
            username = revisions[0]['username']
            try:
                user = resource.get_resource('/users/%s' % username)
            except LookupError:
                return username
            return user.get_title()
        elif column == 'size':
            # Size
            return item.get_human_size()
        elif column == 'workflow_state':
            # The workflow state
            if not isinstance(item, WorkflowAware):
                return None
            statename = item.get_statename()
            state = item.get_state()
            msg = state['title'].gettext().encode('utf-8')
            path = context.get_link(item)
            # TODO Include the template in the base table
            state = ('<a href="%s/;edit_state" class="workflow">'
                     '<strong class="wf_%s">%s</strong>'
                     '</a>') % (path, statename, msg)
            return XMLParser(state)


    table_actions = [
        RemoveButton, RenameButton, CopyButton, CutButton, PasteButton,
        PublishButton]


    #######################################################################
    # Form Actions
    #######################################################################
    def action_remove(self, resource, context, form):
        ids = form['ids']

        # Clean the copy cookie if needed
        cut, paths = context.get_cookie('ikaaro_cp', type=CopyCookie)

        # Remove resources
        removed = []
        referenced = []
        not_removed = []
        user = context.user
        abspath = resource.get_abspath()

        # We sort and reverse ids in order to
        # remove the childs then their parents
        ids.sort()
        ids.reverse()
        for name in ids:
            child = resource.get_resource(name)
            ac = child.get_access_control()
            if ac.is_allowed_to_remove(user, child):
                # Remove resource
                try:
                    resource.del_resource(name)
                except ConsistencyError:
                    referenced.append(name)
                    continue
                removed.append(name)
                # Clean cookie
                if str(abspath.resolve2(name)) in paths:
                    context.del_cookie('ikaaro_cp')
                    paths = []
            else:
                not_removed.append(name)

        message = []
        if removed:
            resources = ', '.join(removed)
            msg = messages.MSG_RESOURCES_REMOVED(resources=resources)
            message.append(msg)
        if referenced:
            resources = ', '.join(referenced)
            msg = messages.MSG_RESOURCES_REFERENCED(resources=resources)
            message.append(msg)
        if not_removed:
            resources = ', '.join(not_removed)
            msg = messages.MSG_RESOURCES_NOT_REMOVED(resources=resources)
            message.append(msg)
        if not removed and not referenced and not not_removed:
            message.append(messages.MSG_NONE_REMOVED)
        context.message = message


    def action_rename(self, resource, context, form):
        ids = form['ids']
        # Filter names which the authenticated user is not allowed to move
        ac = resource.get_access_control()
        user = context.user
        paths = [ x for x in ids
                  if ac.is_allowed_to_move(user, resource.get_resource(x)) ]

        # Check input data
        if not paths:
            context.message = messages.MSG_NONE_SELECTED
            return

        # FIXME Hack to get rename working. The current user interface forces
        # the rename_form to be called as a form action, hence with the POST
        # method, but it should be a GET method. Maybe it will be solved after
        # the needed folder browse overhaul.
        ids_list = '&'.join([ 'ids=%s' % x for x in paths ])
        uri = '%s/;rename?%s' % (context.get_link(resource), ids_list)
        return get_reference(uri)


    def action_copy(self, resource, context, form):
        ids = form['ids']
        # Filter names which the authenticated user is not allowed to copy
        ac = resource.get_access_control()
        user = context.user
        names = [ x for x in ids
                  if ac.is_allowed_to_copy(user, resource.get_resource(x)) ]

        # Check input data
        if not names:
            message = messages.MSG_NONE_SELECTED
            return

        abspath = resource.get_abspath()
        cp = (False, [ str(abspath.resolve2(x)) for x in names ])
        cp = CopyCookie.encode(cp)
        context.set_cookie('ikaaro_cp', cp, path='/')
        # Ok
        context.message = messages.MSG_COPIED


    def action_cut(self, resource, context, form):
        ids = form['ids']
        # Filter names which the authenticated user is not allowed to move
        ac = resource.get_access_control()
        user = context.user
        names = [ x for x in ids
                  if ac.is_allowed_to_move(user, resource.get_resource(x)) ]

        # Check input data
        if not names:
            context.message = messages.MSG_NONE_SELECTED
            return

        abspath = resource.get_abspath()
        cp = (True, [ str(abspath.resolve2(x)) for x in names ])
        cp = CopyCookie.encode(cp)
        context.set_cookie('ikaaro_cp', cp, path='/')

        context.message = messages.MSG_CUT


    action_paste_schema = {}
    def action_paste(self, resource, context, form):
        # Check there is something to paste
        cut, paths = context.get_cookie('ikaaro_cp', type=CopyCookie)
        if len(paths) == 0:
            context.message = messages.MSG_NO_PASTE
            return

        # Paste
        target = resource
        allowed_types = tuple(target.get_document_types())
        pasted = []
        not_allowed = []
        for path in paths:
            # Check the resource actually exists
            try:
                resource = target.get_resource(path)
            except LookupError:
                continue
            if not isinstance(resource, allowed_types):
                not_allowed.append(resource.name)
                continue

            # If cut&paste in the same place, do nothing
            if cut is True:
                source = resource.parent
                if target.get_canonical_path() == source.get_canonical_path():
                    pasted.append(resource.name)
                    continue

            name = generate_name(resource.name, target.get_names(), '_copy_')
            pasted.append(name)
            if cut is True:
                # Cut&Paste
                target.move_resource(path, name)
            else:
                # Copy&Paste
                target.copy_resource(path, name)
                # Fix state
                resource = target.get_resource(name)
                if isinstance(resource, WorkflowAware):
                    metadata = resource.metadata
                    metadata.set_property('state',
                                          resource.workflow.initstate)

        # Cut, clean cookie
        if cut is True:
            context.del_cookie('ikaaro_cp')

        message = []
        if pasted:
            resources = ', '.join(pasted)
            message.append(messages.MSG_RESOURCES_PASTED(resources=resources))
        if not_allowed:
            resources = ', '.join(not_allowed)
            msg = messages.MSG_RESOURCES_NOT_PASTED(resources=resources)
            message.append(msg)

        context.message = message


    def action_publish(self, resource, context, form):
        resources = [ resource.get_resource(id) for id in form['ids'] ]
        user = context.user
        transition = 'publish'
        # Check there is at least one item we can publish
        ac = resource.get_access_control()
        allowed = [ x for x in resources
                    if ac.is_allowed_to_trans(user, x, transition) ]
        if not allowed:
            context.message = messages.MSG_NONE_ALLOWED
            return

        # Publish
        for item in allowed:
            if item.get_statename() == 'public':
                continue
            # Update workflow history
            property = {'date': datetime.now(),
                        'user': user.name,
                        'name': transition,
                        'comments': u""}
            item.set_property('wf_transition', property)
            item.do_trans(transition)

        # Ok
        context.message = messages.MSG_PUBLISHED



class Folder_PreviewContent(Folder_BrowseContent):

    title = MSG(u'Preview Content')
    context_menus = Folder_BrowseContent.context_menus + [ZoomMenu()]
    # Table
    table_template = '/ui/folder/browse_image.xml'


    def get_query_schema(self):
        # Define a huge batch limit, and the image size parameter
        return merge_dicts(Folder_BrowseContent.get_query_schema(self),
                           batch_size=Integer(default=0),
                           size=Integer(default=128),
                           width=String,
                           height=String)


    def get_items(self, resource, context):
        # Show only images
        query = OrQuery(PhraseQuery('is_image', '1'),
                        PhraseQuery('format', 'folder'))
        return Folder_BrowseContent.get_items(self, resource, context, query)


    def get_table_head(self, resource, context, items, actions=None):
        # Get from the query
        query = context.query
        sort_by = query['sort_by']
        reverse = query['reverse']

        columns = self._get_table_columns(resource, context)
        columns_ns = []
        for name, title, sortable in columns:
            if name == 'checkbox':
                # Type: checkbox
                if  actions:
                    columns_ns.append({'is_checkbox': True})
            elif title is None:
                # Type: nothing
                continue
            elif not sortable:
                # Type: nothing or not sortable
                columns_ns.append({
                    'is_checkbox': False,
                    'title': title,
                    'href': None})
            else:
                # Type: normal
                kw = {'sort_by': name}
                if name == sort_by:
                    col_reverse = (not reverse)
                    order = 'up' if reverse else 'down'
                else:
                    col_reverse = False
                    order = 'none'
                kw['reverse'] = Boolean.encode(col_reverse)
                columns_ns.append({
                    'is_checkbox': False,
                    'title': title,
                    'order': order,
                    'href': context.uri.replace(**kw),
                    })


    def get_table_namespace(self, resource, context, items):
        context.styles.append('/ui/gallery/style.css')
        context.scripts.append('/ui/gallery/javascript.js')

        # Get from the query
        query = context.query
        width = query['width']
        height = query['height']

        # (0) Zoom
        current_size = query['size']
        min_size = resource.MIN_SIZE
        max_size = resource.MAX_SIZE
        current_size = max(min_size, min(current_size, max_size))

        # (1) Actions (submit buttons)
        ac = resource.get_access_control()
        actions = []
        for button in self.get_table_actions(resource, context):
            if button.show(resource, context, items) is False:
                continue
            if button.confirm:
                confirm = button.confirm.gettext().encode('utf_8')
                onclick = 'return confirm("%s");' % confirm
            else:
                onclick = None
            actions.append(
                {'name': button.name,
                 'value': button.title,
                 'class': button.css,
                 'onclick': onclick})

        # (2) Table Head: columns
        table_head = self.get_table_head(resource, context, items, actions)

        # (3) Table Body: rows
        columns = self._get_table_columns(resource, context)
        rows = []
        for item in items:
            row = {'checkbox': False,
                   # These are required for internal use
                   'title_or_name': item.get_title(),
                   'workflow_statename': None}
            # XXX Already hard-coded in the catalog search
            row['is_folder'] = (item.class_id == 'folder')
            if isinstance(item, WorkflowAware):
                row['workflow_statename'] = item.get_statename()
            for name, title, sortable in columns:
                value = self.get_item_value(resource, context, item, name)
                if value is None:
                    continue
                elif name == 'checkbox':
                    if actions:
                        value, checked = value
                        row['checkbox'] = True
                        row['id'] = value
                        row['checked'] = checked
                elif name == 'name':
                    if type(value) is tuple:
                        value, href = value
                        href = get_reference(href)
                        if row['is_folder']:
                            href = href.resolve2(';preview_content')
                        href = href.replace(size=current_size, width=width,
                                height=height)
                        href = str(href)
                    else:
                        href = None
                    row['name'] = value
                    row['href'] = href
                else:
                    row[name] = value
            rows.append(row)

        widths = ", ".join([repr(o['name']) for o in ImageWidth.get_options()
                            if o['name'].strip()])

        return {'size': current_size,
                'width': width,
                'height': height,
                'widths': widths,
                'css': self.table_css,
                'columns': table_head,
                'rows': rows,
                'actions': actions}



class Folder_LastChanges(Folder_BrowseContent):

    title = MSG(u"Last Changes")

    def get_query_schema(self):
        schema = Folder_BrowseContent.get_query_schema(self)

        # Override the default values
        schema['reverse'] = Boolean(default=True)
        schema['sort_by'] = String(default='mtime')
        schema['search_subfolders'] = Boolean(default=True)
        return schema


    def get_items(self, resource, context):
        # Show only version aware resources
        query = PhraseQuery('is_version_aware', '1')
        return Folder_BrowseContent.get_items(self, resource, context, query)



class Folder_Orphans(Folder_BrowseContent):
    """Orphans are files not referenced in another resource of the database.
    It extends the concept of "orphans pages" from the wiki to all file-like
    resources.

    Orphans folders generally don't make sense because they serve as
    containers. TODO or list empty folders?
    """

    access = 'is_allowed_to_view'
    title = MSG(u"Orphans")
    icon = 'orphans.png'
    description = MSG(u"Show resources not linked from anywhere.")


    def get_items(self, resource, context):
        # Make the base search
        items = Folder_BrowseContent.get_items(self, resource, context)

        # Find out the orphans
        root = context.root
        orphans = []
        for item in items.get_documents():
            query = PhraseQuery('links', item.abspath)
            results = root.search(query)
            if len(results) == 0:
                orphans.append(item)

        # Transform back the items found in a SearchResults object.
        # FIXME This is required by 'get_item_value', we should change that,
        # for better performance.
        args = [ PhraseQuery('abspath', x.abspath) for x in orphans ]
        query = OrQuery(*args)
        items = root.search(query)

        # Ok
        return items



class Folder_Thumbnail(BaseView):

    access = True

    default_icon = '/ui/gallery/folder.png'

    def get_mtime(self, resource):
        return resource.get_mtime()


    def GET(self, resource, context):
        from file import Image

        width = context.get_form_value('width', type=Integer, default=48)
        height = context.get_form_value('height', type=Integer, default=48)
        size = (width, height)
        data = None
        format = "jpeg"

        # Choose an image to illustrate
        default_icon = resource.get_resource(self.default_icon)
        if PILImage is None:
            # Full size but better than nothing
            data = default_icon.to_str()
            format = 'png'
        else:
            # Find the first accessible image
            user = context.user
            ac = resource.get_access_control()
            for image in resource.search_resources(cls=Image):
                # Search public image safe for all
                if ac.is_allowed_to_view(user, image):
                    data, format = image.handler.get_thumbnail(width, height)
                    break
            else:
                # Default icon for empty or inaccessible folders
                data, format = default_icon.get_thumbnail(width, height)

        # XXX Don't cache nothing here
        # The image thumbnail was cached in the image handler
        # The folder thumbnail was cached in the folder handler
        # Accessible images depend on too many parameters

        response = context.response
        response.set_header('Content-Type', 'image/%s' % format)
        return data



class GoToSpecificDocument(BaseView):

    access = 'is_allowed_to_view'
    title = MSG(u'Front Page')
    icon = 'view.png'
    specific_document = 'FrontPage'


    def get_specific_document(self, resource, context):
        return self.specific_document


    def GET(self, resource, context):
        specific_document = self.get_specific_document(resource, context)
        goto = '%s/%s' % (context.get_link(resource), specific_document)
        goto = get_reference(goto)

        # Keep the message
        if context.has_form_value('message'):
            message = context.get_form_value('message')
            goto = goto.replace(message=message)

        return goto




#!/usr/bin/env python
# -*- coding: UTF-8 -*-
# Copyright (C) 2007 Hervé Cauwelier <herve@itaapy.com>
# Copyright (C) 2007-2008 Juan David Ibáñez Palomar <jdavid@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from the Standard Library
from optparse import OptionParser
import sys
from time import time

# Import from itools
import itools
from itools.handlers import Database
from itools.utils import vmsize
from itools import vfs
from itools.xapian import make_catalog, CatalogAware
from itools.web import Context
from itools.i18n.accept import AcceptLanguage
from itools.http import Request

# Import from ikaaro
from ikaaro.server import ask_confirmation
from ikaaro.server import Server, get_pid


def update_catalog(parser, options, target):
    # Check for database consistency
    if vfs.exists('%s/database.commit' % target):
        print 'The database is not in a consistent state, to fix it up type:'
        print
        print '    $ icms-restore.py <instance>'
        print
        return

    # Check the server is not running
    pid = get_pid(target)
    if pid is not None:
        print 'The server is running. To update the catalog first stop the'
        print 'server.'
        return

    # Ask
    message = 'Update the catalog (y/N)? '
    if ask_confirmation(message, options.confirm) is False:
        return

    # Remove the old catalog and create a new one
    catalog_path = '%s/catalog' % target
    if vfs.exists(catalog_path):
        vfs.remove(catalog_path)
    catalog = make_catalog(catalog_path)

    # Get the root
    server = Server(target, read_only=True)
    server.database.set_use_cache(False)
    root = server.root

    # Build a fake context
    context = Context(Request())
    context.accept_language = AcceptLanguage()
    context.uri = None
    server.init_context(context)

    # Update
    t0, v0 = time(), vmsize()
    doc_n = 0
    for obj in root.traverse_resources():
        if not isinstance(obj, CatalogAware):
            continue
        print doc_n, obj.get_abspath()
        doc_n += 1
        catalog.index_document(obj)
        # Free Memory
        del obj
    # Update / Report
    t1, v1 = time(), vmsize()
    v = (v1 - v0)/1024
    print '[Update] Time: %.02f seconds. Memory: %s Kb' % (t1 - t0, v)

    # Commit
    print '[Commit]',
    sys.stdout.flush()
    catalog.save_changes()
    # Commit / Report
    t2, v2 = time(), vmsize()
    v = (v2 - v1)/1024
    print 'Time: %.02f seconds. Memory: %s Kb' % (t2 - t1, v)



if __name__ == '__main__':
    # The command line parser
    usage = '%prog [OPTIONS] TARGET'
    version = 'itools %s' % itools.__version__
    description = (
        'Rebuilds the catalog: first removes and creates a new empty one;'
        ' then traverses and indexes all resources in the database.')
    parser = OptionParser(usage, version=version, description=description)
    parser.add_option(
        '-y', '--yes', action='store_true', dest='confirm',
        help="start the update without asking confirmation")

    options, args = parser.parse_args()
    if len(args) != 1:
        parser.error('incorrect number of arguments')

    target = args[0]

    # Action!
    update_catalog(parser, options, target)



# -*- coding: UTF-8 -*-
# Copyright (C) 2005-2008 Juan David Ibáñez Palomar <jdavid@itaapy.com>
# Copyright (C) 2006-2008 Hervé Cauwelier <herve@itaapy.com>
# Copyright (C) 2007 Sylvain Taverne <sylvain@itaapy.com>
# Copyright (C) 2007-2008 Henry Obein <henry@itaapy.com>
# Copyright (C) 2008 Matthieu France <matthieu@itaapy.com>
# Copyright (C) 2008 Nicolas Deram <nicolas@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from the Standard Library
from operator import itemgetter
from urllib import quote

# Import from itools
from itools.datatypes import String, Unicode
from itools.gettext import MSG
from itools.handlers import checkid
from itools.i18n import get_language_name
from itools.handlers import merge_dicts
from itools.uri import Path, get_reference
from itools.vfs import FileName
from itools.web import get_context, BaseView, STLForm, INFO, ERROR

# Import from ikaaro
from datatypes import FileDataType
from forms import AutoForm, title_widget, description_widget, subject_widget
from forms import TextWidget
import messages
from registry import get_resource_class
from utils import get_parameters, reduce_string
from views import NewInstanceForm, ContextMenu



class EditLanguageMenu(ContextMenu):

    title = MSG(u'Edit Language')

    def get_items(self, resource, context):
        content_language = resource.get_content_language(context)

        site_root = resource.get_site_root()
        languages = site_root.get_property('website_languages')
        return [
            {'title': get_language_name(x),
             'href': context.uri.replace(content_language=x),
             'class': 'nav_active' if (x == content_language) else None}
            for x in languages ]



class AddResourceMenu(ContextMenu):

    title = MSG(u'Add Resource')

    def get_items(self, resource, context):
        base = '%s/;new_resource' % context.get_link(resource)
        document_types = resource.get_document_types()
        return [
            {'src': '/ui/' + cls.class_icon16,
             'title': cls.class_title.gettext(),
             'href': '%s?type=%s' % (base, quote(cls.class_id))}
            for cls in document_types ]


class DBResource_NewInstance(NewInstanceForm, AutoForm):

    access = 'is_allowed_to_add'
    query_schema = {
        'type': String}
    schema = {
        'name': String,
        'title': Unicode}
    widgets = [
        title_widget,
        TextWidget('name', title=MSG(u'Name'), default='')]
    submit_value = MSG(u'Add')
    context_menus = [AddResourceMenu()]


    def get_title(self, context):
        if self.title is not None:
            return self.title
        type = context.get_query_value('type')
        if not type:
            return MSG(u'Add resource').gettext()
        cls = get_resource_class(type)
        class_title = cls.class_title.gettext()
        title = MSG(u'Add $class_title')
        return title.gettext(class_title=class_title)


    def get_new_resource_name(self, form):
        # If the name is not explicitly given, use the title
        name = form['name']
        title = form['title'].strip()
        if name is None:
            return title
        return name or title


    def action(self, resource, context, form):
        name = form['name']
        title = form['title']

        # Create the resource
        class_id = context.query['type']
        cls = get_resource_class(class_id)
        child = cls.make_resource(cls, resource, name)
        # The metadata
        metadata = child.metadata
        language = resource.get_content_language(context)
        metadata.set_property('title', title, language=language)

        goto = './%s/' % name
        return context.come_back(messages.MSG_NEW_RESOURCE, goto=goto)



class DBResource_Edit(AutoForm):

    access = 'is_allowed_to_edit'
    title = MSG(u'Edit')
    icon = 'metadata.png'
    context_menus = [EditLanguageMenu()]

    schema = {
        'title': Unicode,
        'description': Unicode,
        'subject': Unicode}
    widgets = [title_widget, description_widget, subject_widget]


    def get_value(self, resource, context, name, datatype):
        language = resource.get_content_language(context)
        return resource.get_property(name, language=language)


    def action(self, resource, context, form):
        title = form['title']
        description = form['description']
        subject = form['subject']
        language = resource.get_content_language(context)
        resource.set_property('title', title, language=language)
        resource.set_property('description', description, language=language)
        resource.set_property('subject', subject, language=language)
        # Ok
        context.message = messages.MSG_CHANGES_SAVED



###########################################################################
# Interface to add images from the TinyMCE editor
###########################################################################
class Breadcrumb(object):
    """Instances of this class will be used as namespaces for STL templates.
    The built namespace contains the breadcrumb, that is to say, the path from
    the tree root to another tree node, and the content of that node.
    """

    def __init__(self, filter_types=None, root=None, start=None,
            icon_size=16):
        """The 'start' must be a handler, 'filter_type' must be a handler
        class.
        """
        from file import Image
        from folder import Folder
        from resource_ import DBResource

        if filter_types is None:
            filter_types = (DBResource,)

        context = get_context()
        here = context.resource
        if root is None:
            root = here.get_site_root()
        if start is None:
            start = root

        # Get the query parameters
        parameters = get_parameters('bc', id=None, target=None)
        id = parameters['id']
        # Get the target folder
        target_path = parameters['target']
        if target_path is None:
            if isinstance(start, Folder):
                target = start
            else:
                target = start.parent
        else:
            target = root.get_resource(target_path)
        self.target_path = str(target.get_abspath())

        # Resource to link
        item = context.get_form_value('item', default='')
        if item == '':
            item = '.'
        self.item = item

        # The breadcrumb
        breadcrumb = []
        node = target
        while node is not root.parent:
            url = context.uri.replace(bc_target=str(root.get_pathto(node)))
            title = node.get_title()
            short_title = reduce_string(title, 12, 40)
            quoted_title = short_title.replace("'", "\\'")
            breadcrumb.insert(0, {'name': node.name,
                                  'title': title,
                                  'short_title': short_title,
                                  'quoted_title': quoted_title,
                                  'url': url})
            node = node.parent
        self.path = breadcrumb

        # Content
        items = []
        self.is_submit = False
        user = context.user
        filters = (Folder,) + filter_types
        for resource in target.search_resources(cls=filters):
            ac = resource.get_access_control()
            if not ac.is_allowed_to_view(user, resource):
                continue
            path = here.get_pathto(resource)
            bc_target = str(root.get_pathto(resource))
            url = context.uri.replace(bc_target=bc_target)

            self.is_submit = True
            # Calculate path
            is_image = isinstance(resource, Image)
            if is_image:
                path_to_icon = ";thumb?width=%s&height=%s" % (icon_size,
                                                              icon_size)
            else:
                path_to_icon = resource.get_resource_icon(icon_size)
            if path:
                path_to_resource = Path(str(path) + '/')
                path_to_icon = path_to_resource.resolve(path_to_icon)
            title = resource.get_title()
            short_title = reduce_string(title, 12, 40)
            quoted_title = short_title.replace("'", "\\'")
            items.append({'name': resource.name,
                          'title': title,
                          'short_title': short_title,
                          'quoted_title': quoted_title,
                          'is_folder': isinstance(resource, Folder),
                          'is_image': is_image,
                          'is_selectable': True,
                          'path': path,
                          'url': url,
                          'icon': path_to_icon,
                          'type': type(resource),
                          'item_type': resource.handler.get_mimetype()})

        items.sort(key=itemgetter('is_folder'), reverse=True)
        self.items = items

        # Avoid general template
        response = context.response
        response.set_header('Content-Type', 'text/html; charset=UTF-8')



class DBResource_AddBase(STLForm):
    """Base class for 'DBResource_AddImage' and 'DBResource_AddLink' (used
    by the Web Page editor).
    """

    access = 'is_allowed_to_edit'

    element_to_add = None

    configuration = {}

    schema = {
        'target_path': String(mandatory=True),
        'target_id': String(default=None),
        'mode': String(mandatory=True),
      }

    styles = ['/ui/bo.css',
              '/ui/aruni/style.css']

    base_scripts = ['/ui/jquery.js',
                    '/ui/javascript.js']


    scripts = {'wiki': ['/ui/wiki/javascript.js'],
               'tiny_mce': ['/ui/tiny_mce/javascript.js',
                            '/ui/tiny_mce/tiny_mce_src.js',
                            '/ui/tiny_mce/tiny_mce_popup.js'],
               'input': []}


    action_upload_schema = merge_dicts(schema,
                                       file=FileDataType(mandatory=True))


    additional_javascript = """
          function select_element(type, value, caption) {
            window.opener.$("#%s").val(value);
            window.close();
          }
          """


    def get_filter_types(self):
        from file import File
        return (File,)


    def get_namespace(self, resource, context):
        from file import File
        # Get some informations
        mode = context.get_form_value('mode')
        # For the breadcrumb
        filter_types = self.get_filter_types()
        if isinstance(resource, File):
            start = resource.parent
        else:
            start = resource
        # Construct namespace
        namespace = self.configuration
        namespace.update({
            'additional_javascript': self.get_additional_javascript(context),
            'bc': Breadcrumb(filter_types=filter_types, start=start,
                             icon_size=48),
            'element_to_add': self.element_to_add,
            'target_id': context.get_form_value('target_id'),
            'message': context.message,
            'mode': mode,
            'resource_action': self.get_resource_action(context),
            'styles': self.styles,
            'scripts': self.get_scripts(mode)})
        return namespace


    def get_scripts(self, mode):
        if mode is None:
            return self.base_scripts
        return self.base_scripts + self.scripts[mode]


    def get_additional_javascript(self, context):
        mode = context.get_form_value('mode')
        if mode!='input':
            return ''
        target_id = context.get_form_value('target_id')
        return self.additional_javascript % target_id


    def action_upload(self, resource, context, form):
        filename, mimetype, body = form['file']
        name, type, language = FileName.decode(filename)

        # Check the filename is good
        name = checkid(name)
        if name is None:
            context.message = messages.MSG_BAD_NAME
            return

        # Get the container
        container = context.root.get_resource(form['target_path'])

        # Check the name is free
        if container.has_resource(name):
            context.message = messages.MSG_NAME_CLASH
            return

        # Check it is of the expected type
        filter_types = self.get_filter_types()
        cls = get_resource_class(mimetype)
        is_compatible = False
        for filter_type in filter_types:
            if issubclass(cls, filter_type):
                is_compatible = True
                break
        if is_compatible is False:
            class_ids = ', '.join([x.class_id for x in filter_types])
            context.message = ERROR(u'The given file is none of the types '
                                    u'$class_ids.', class_ids=class_ids)
            return

        # Add the image to the resource
        cls.make_resource(cls, container, name, body, format=mimetype,
                          filename=filename, extension=type)
        # Get resource path
        child = container.get_resource(name)
        path = resource.get_pathto(child)
        # Add an action to the resource
        action = self.get_resource_action(context)
        if action:
            path = path.resolve2('.%s' % action)
        # Return javascript
        mode = form['mode']
        context.scripts.extend(self.get_scripts(mode))
        return self.get_javascript_return(context, path)


    def get_javascript_return(self, context, path):
        return """
            <script type="text/javascript">
                %s
                select_element('%s', '%s', '');
            </script>""" % (self.get_additional_javascript(context),
                            self.element_to_add, path)


    def get_resource_action(self, context):
        return ''



class DBResource_AddImage(DBResource_AddBase):

    element_to_add = 'image'

    template = '/ui/html/addimage.xml'

    configuration = {'show_browse': True,
                     'show_upload': True}


    def get_filter_types(self):
        from file import Image
        return (Image,)


    def get_resource_action(self, context):
        mode = context.get_form_value('mode')
        if mode=='tiny_mce':
            return '/;download'
        return DBResource_AddBase.get_resource_action(self, context)



class DBResource_AddLink(DBResource_AddBase):

    template = '/ui/html/addlink.xml'

    element_to_add = 'link'

    action_add_resource_schema = merge_dicts(DBResource_AddImage.schema,
                                             title=String(mandatory=True))

    configuration = {'show_browse': True,
                     'show_external': True,
                     'show_insert': True,
                     'show_upload': True}


    def action_add_resource(self, resource, context, form):
        mode = form['mode']
        name = checkid(form['title'])
        # Check name validity
        if name is None:
            context.message = MSG(u"Invalid title.")
            return
        # Get the container
        root = context.root
        container = root.get_resource(context.get_form_value('target_path'))
        # Check the name is free
        if container.has_resource(name):
            context.message = messages.MSG_NAME_CLASH
            return
        # Get the type of resource to add
        cls = self.get_page_type(mode)
        # Create the resource
        child = cls.make_resource(cls, container, name)
        path = context.resource.get_pathto(child)
        context.scripts.extend(self.get_scripts(mode))
        return self.get_javascript_return(context, path)


    def get_page_type(self, mode):
        """Return the type of page to add corresponding to the mode
        """
        if mode=='tiny_mce':
            from wiki.page import WikiPage
            return WikiPage
        elif mode=='wiki':
            from ikaaro.html import WebPage
            return WebPage
        else:
            raise ValueError, 'Incorrect mode %s' % mode


###########################################################################
# Views / Login, Logout
###########################################################################

class LoginView(STLForm):

    access = True
    title = MSG(u'Login')
    template = '/ui/base/login.xml'
    schema = {
        'username': Unicode(mandatory=True),
        'password': String(mandatory=True),
    }


    def get_namespace(self, resource, context):
        return {
            'username': context.get_form_value('username'),
        }


    def action(self, resource, context, form):
        email = form['username']
        password = form['password']

        # Check the user exists
        root = context.root
        user = root.get_user_from_login(email)
        if user is None:
            message = ERROR(u'The user "$username" does not exist.',
                            username=email)
            context.message = message
            return

        # Check the password is right
        if not user.authenticate(password):
            context.message = ERROR(u'The password is wrong.')
            return

        # Set cookie
        user.set_auth_cookie(context, password)

        # Set context
        context.user = user

        # Come back
        goto = None
        referrer = context.request.referrer
        if referrer:
            if not referrer.path:
                goto = referrer
            else:
                params = referrer.path[-1].params
                if not params:
                    goto = referrer
                elif params[0] != 'login':
                    goto = referrer
        if goto is None:
            goto = get_reference('./')

        return context.come_back(INFO(u"Welcome!"), goto)



class LogoutView(BaseView):
    """Logs out of the application.
    """

    access = True


    def GET(self, resource, context):
        # Log-out
        context.del_cookie('__ac')
        context.user = None

        message = INFO(u'You Are Now Logged out.')
        return context.come_back(message, goto='./')



# -*- coding: UTF-8 -*-
# Copyright (C) 2006-2008 Hervé Cauwelier <herve@itaapy.com>
# Copyright (C) 2006-2008 Nicolas Deram <nicolas@itaapy.com>
# Copyright (C) 2007-2008 Juan David Ibáñez Palomar <jdavid@itaapy.com>
# Copyright (C) 2007-2008 Sylvain Taverne <sylvain@itaapy.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import from Standard Library
from datetime import time

# Import from itools
from itools.datatypes import DataType, Date
from itools.gettext import MSG
from itools.ical import iCalendar, icalendarTable
from itools.ical import Record
from itools.web import get_context

# Import from ikaaro
from calendar_views import Calendar_Upload, Calendar_Download
from calendar_views import AddEventForm, EditEventForm
from calendar_views import MonthlyView, TimetablesForm, WeeklyView, DailyView
from file_views import File_View
from folder import Folder
from registry import register_resource_class
from resource_ import DBResource
from table import Table
from text import Text


class Timetables(DataType):
    """Timetables are tuples of time objects (start, end) used by cms.ical.

    Example with 3 timetables as saved into metadata:
        (8,0),(10,0);(10,0),(12,0);(15,30),(17,30)

    Decoded value are:
        [(time(8,0), time(10,0)), (time(10,0), time(12, 0)),
         (time(15,30), time(17, 30))]
    """

    default = ()

    @staticmethod
    def decode(value):
        if not value:
            return ()
        timetables = []
        for timetable in value.strip().split(';'):
            start, end = timetable[1:-1].split('),(')
            hours, minutes = start.split(',')
            hours, minutes = int(hours), int(minutes)
            start = time(hours, minutes)
            hours, minutes = end.split(',')
            hours, minutes = int(hours), int(minutes)
            end = time(hours, minutes)
            timetables.append((start, end))
        return tuple(timetables)


    @staticmethod
    def encode(value):
        timetables = []
        for start, end in value:
            start = '(' + str(start.hour) + ',' + str(start.minute) + ')'
            end = '(' + str(end.hour) + ',' + str(end.minute) + ')'
            timetables.append(start + ',' + end)
        return ';'.join(timetables)



class CalendarBase(DBResource):

    class_title = MSG(u'Calendar')
    class_description = MSG(u'Schedule your time with calendar files.')
    class_icon16 = 'icons/16x16/icalendar.png'
    class_icon48 = 'icons/48x48/icalendar.png'
    class_views = ['monthly_view', 'weekly_view', 'daily_view',
                   'edit_timetables', 'upload', 'download_form']


    timetables = [((7,0),(8,0)), ((8,0),(9,0)), ((9,0),(10,0)),
                  ((10,0),(11,0)), ((11,0),(12,0)), ((12,0),(13,0)),
                  ((13,0),(14,0)), ((14,0),(15,0)), ((15,0),(16,0)),
                  ((16,0),(17,0)), ((17,0),(18,0)), ((18,0),(19,0)),
                  ((19,0),(20,0)), ((20,0),(21,0))]


    def get_calendars(self):
        return [self]


    def get_action_url(self, **kw):
        if 'day' in kw:
            return ';add_event?date=%s' % Date.encode(kw['day'])
        if 'id' in kw:
            return ';edit_event?id=%s' % kw['id']

        return None


    # Test if user in context is the organizer of a given event (or is admin)
    def is_organizer_or_admin(self, context, event):
        if self.get_access_control().is_admin(context.user, self):
            return True
        if event:
            organizer = event.get_property('ORGANIZER')
            user_path = str(context.user.get_abspath())
            return organizer and user_path == organizer.value
        ac = self.parent.get_access_control()
        return ac.is_allowed_to_edit(context.user, self.parent)


    def get_timetables(self):
        """Build a list of timetables represented as tuples(start, end).
        Data are taken from metadata or from class value.

        Example of metadata:
          <timetables>(8,0),(10,0);(10,30),(12,0);(13,30),(17,30)</timetables>
        """
        if self.has_property('timetables'):
            return self.get_property('timetables')

        # From class value
        timetables = []
        for index, (start, end) in enumerate(self.timetables):
            timetables.append((time(start[0], start[1]), time(end[0], end[1])))
        return timetables


    def get_events_to_display(self, start, end):
        file = self.handler
        events = []
        for event in file.search_events_in_range(start, end, sortby='date'):
            e_dtstart = event.get_property('DTSTART').value
            events.append((self.name, e_dtstart, event))
        events.sort(lambda x, y : cmp(x[1], y[1]))
        return {self.name: 0}, events


    #######################################################################
    # Views
    #######################################################################
    monthly_view = MonthlyView()
    weekly_view = WeeklyView()
    daily_view = DailyView()
    add_event = AddEventForm()
    edit_event = EditEventForm()
    edit_timetables = TimetablesForm()
    download = Calendar_Download()
    upload = Calendar_Upload()
    download_form = File_View()



class CalendarTable(CalendarBase, Table):

    class_id = 'calendarTable'
    class_version = '20071216'
    class_handler = icalendarTable
    record_class = Record


    def get_record(self, id):
        id = int(id)
        return self.handler.get_record(id)


    def add_record(self, type, properties):
        properties['type'] = type
        # Reindex the resource
        get_context().server.change_resource(self)
        return self.handler.add_record(properties)


    def update_record(self, id, properties):
        id = int(id)
        self.handler.update_record(id, **properties)
        # Reindex the resource
        get_context().server.change_resource(self)


    def _remove_event(self, uid):
        self.handler.del_record(int(uid))
        # Reindex the resource
        get_context().server.change_resource(self)


    @classmethod
    def get_metadata_schema(cls):
        schema = Table.get_metadata_schema()
        schema['timetables'] = Timetables
        return schema


    # Use edit_event instead
    edit_record = None



class Calendar(CalendarBase, Text):

    class_id = 'text/calendar'
    class_version = '20071216'
    class_handler = iCalendar


    def get_record(self, id):
        return self.handler.get_record(id)


    def add_record(self, type, properties):
        # Reindex the resource
        get_context().server.change_resource(self)
        return self.handler.add_component(type, **properties)


    def update_record(self, id, properties):
        self.handler.update_component(id, **properties)
        # Reindex the resource
        get_context().server.change_resource(self)


    def _remove_event(self, uid):
        self.handler.remove(uid)
        # Reindex the resource
        get_context().server.change_resource(self)


    @classmethod
    def get_metadata_schema(cls):
        schema = Text.get_metadata_schema()
        schema['timetables'] = Timetables
        return schema



class CalendarContainer(CalendarBase):

    @classmethod
    def get_metadata_schema(cls):
        return {'timetables': Timetables}


    def get_calendars(self, types=None):
        """List of sources from which taking events.
        """
        if types is None:
            types = (Calendar, CalendarTable)

        if isinstance(self, Folder):
            calendars = self.search_resources(cls=types)
            return list(calendars)
        return [self]


    # Views
    download = None
    upload = None



###########################################################################
# Register
###########################################################################
register_resource_class(CalendarTable)
register_resource_class(Calendar)
